# Copyright (C) 2021 Istituto Italiano di Tecnologia (IIT). All rights reserved.
# This software may be modified and distributed under the terms of the
# GNU Lesser General Public License v2.1 or any later version.

import pytest

pytestmark = pytest.mark.scenario

from typing import Tuple

import numpy as np

from scenario import core
from scenario import gazebo as scenario

from ..common import utils
from ..common.utils import default_world_fixture as default_world

# Set the verbosity
scenario.set_verbosity(scenario.Verbosity_debug)


@pytest.mark.parametrize("default_world", [(1.0 / 1_000, 1.0, 1)], indirect=True)
def test_fixed_base(default_world: Tuple[scenario.GazeboSimulator, scenario.World]):

    # Get the simulator and the world
    gazebo, world = default_world

    # Insert a sphere
    pose = core.Pose([0, 0.0, 1.0], [1.0, 0, 0, 0])
    assert world.insert_model_from_string(
        utils.SphereURDF(mass=10.0).urdf(), pose, "sphere"
    )
    assert "sphere" in world.model_names()

    # Get the sphere
    sphere = world.get_model("sphere")
    link = sphere.get_link("sphere").to_gazebo()

    # Get the initial z position
    initial_height = link.position()[2]

    # Apply a vertical force that counterbalances gravity
    assert link.apply_world_force(
        force=-np.array(world.gravity()) * sphere.total_mass(), duration=0.5
    )

    # Run the simulation for a while
    for _ in range(500):
        gazebo.run()
        assert link.position()[2] == pytest.approx(initial_height)

    # Run the simulation for a while
    for _ in range(500):
        gazebo.run()
        assert link.position()[2] < initial_height



import enum
import time
from functools import partial
from typing import List

import gym_ignition
import gym_ignition_environments
import numpy as np
from gym_ignition.context.gazebo import controllers
from gym_ignition.rbd import conversions
from gym_ignition.rbd.idyntree import inverse_kinematics_nlp
from scipy.spatial.transform import Rotation as R

from scenario import core as scenario_core
from scenario import gazebo as scenario_gazebo

# Configure verbosity
scenario_gazebo.set_verbosity(scenario_gazebo.Verbosity_error)

# Configure numpy output
np.set_printoptions(precision=4, suppress=True)


def add_panda_controller(
    panda: gym_ignition_environments.models.panda.Panda, controller_period: float
) -> None:

    # Set the controller period
    assert panda.set_controller_period(period=controller_period)

    # Increase the max effort of the fingers
    panda.get_joint(
        joint_name="panda_finger_joint1"
    ).to_gazebo().set_max_generalized_force(max_force=500.0)
    panda.get_joint(
        joint_name="panda_finger_joint2"
    ).to_gazebo().set_max_generalized_force(max_force=500.0)

    # Insert the ComputedTorqueFixedBase controller
    assert panda.to_gazebo().insert_model_plugin(
        *controllers.ComputedTorqueFixedBase(
            kp=[100.0] * (panda.dofs() - 2) + [10000.0] * 2,
            ki=[0.0] * panda.dofs(),
            kd=[17.5] * (panda.dofs() - 2) + [100.0] * 2,
            urdf=panda.get_model_file(),
            joints=panda.joint_names(),
        ).args()
    )

    # Initialize the controller to the current state
    assert panda.set_joint_position_targets(panda.joint_positions())
    assert panda.set_joint_velocity_targets(panda.joint_velocities())
    assert panda.set_joint_acceleration_targets(panda.joint_accelerations())


def get_panda_ik(
    panda: gym_ignition_environments.models.panda.Panda, optimized_joints: List[str]
) -> inverse_kinematics_nlp.InverseKinematicsNLP:

    # Create IK
    ik = inverse_kinematics_nlp.InverseKinematicsNLP(
        urdf_filename=panda.get_model_file(),
        considered_joints=optimized_joints,
        joint_serialization=panda.joint_names(),
    )

    # Initialize IK
    ik.initialize(
        verbosity=1,
        floating_base=False,
        cost_tolerance=1e-8,
        constraints_tolerance=1e-8,
        base_frame=panda.base_frame(),
    )

    # Set the current configuration
    ik.set_current_robot_configuration(
        base_position=np.array(panda.base_position()),
        base_quaternion=np.array(panda.base_orientation()),
        joint_configuration=np.array(panda.joint_positions()),
    )

    # Add the cartesian target of the end effector
    end_effector = "end_effector_frame"
    ik.add_target(
        frame_name=end_effector,
        target_type=inverse_kinematics_nlp.TargetType.POSE,
        as_constraint=False,
    )

    return ik


def insert_bucket(world: scenario_gazebo.World) -> scenario_gazebo.Model:

    # Insert objects from Fuel
    uri = lambda org, name: f"https://fuel.ignitionrobotics.org/{org}/models/{name}"

    # Download the cube SDF file
    bucket_sdf = scenario_gazebo.get_model_file_from_fuel(
        uri=uri(
            org="GoogleResearch",
            name="Threshold_Basket_Natural_Finish_Fabric_Liner_Small",
        ),
        use_cache=False,
    )

    # Assign a custom name to the model
    model_name = "bucket"

    # Insert the model
    assert world.insert_model(
        bucket_sdf, scenario_core.Pose([0.68, 0, 1.02], [1.0, 0, 0, 1]), model_name
    )

    # Return the model
    return world.get_model(model_name=model_name)


def insert_table(world: scenario_gazebo.World) -> scenario_gazebo.Model:

    # Insert objects from Fuel
    uri = lambda org, name: f"https://fuel.ignitionrobotics.org/{org}/models/{name}"

    # Download the cube SDF file
    bucket_sdf = scenario_gazebo.get_model_file_from_fuel(
        uri=uri(org="OpenRobotics", name="Table"), use_cache=False
    )

    # Assign a custom name to the model
    model_name = "table"

    # Insert the model
    assert world.insert_model(bucket_sdf, scenario_core.Pose_identity(), model_name)

    # Return the model
    return world.get_model(model_name=model_name)


def insert_cube_in_operating_area(
    world: scenario_gazebo.World,
) -> scenario_gazebo.Model:

    # Insert objects from Fuel
    uri = lambda org, name: f"https://fuel.ignitionrobotics.org/{org}/models/{name}"

    # Download the cube SDF file
    cube_sdf = scenario_gazebo.get_model_file_from_fuel(
        uri=uri(org="openrobotics", name="wood cube 5cm"), use_cache=False
    )

    # Sample a random position
    random_position = np.random.uniform(low=[0.2, -0.3, 1.01], high=[0.4, 0.3, 1.01])

    # Get a unique name
    model_name = gym_ignition.utils.scenario.get_unique_model_name(
        world=world, model_name="cube"
    )

    # Insert the model
    assert world.insert_model(
        cube_sdf, scenario_core.Pose(random_position, [1.0, 0, 0, 0]), model_name
    )

    # Return the model
    return world.get_model(model_name=model_name)


def solve_ik(
    target_position: np.ndarray,
    target_orientation: np.ndarray,
    ik: inverse_kinematics_nlp.InverseKinematicsNLP,
) -> np.ndarray:

    quat_xyzw = R.from_euler(seq="y", angles=90, degrees=True).as_quat()

    ik.update_transform_target(
        target_name=ik.get_active_target_names()[0],
        position=target_position,
        quaternion=conversions.Quaternion.to_wxyz(xyzw=quat_xyzw),
    )

    # Run the IK
    ik.solve()

    return ik.get_reduced_solution().joint_configuration


def end_effector_reached(
    position: np.array,
    end_effector_link: scenario_core.Link,
    max_error_pos: float = 0.01,
    max_error_vel: float = 0.5,
    mask: np.ndarray = np.array([1.0, 1.0, 1.0]),
) -> bool:

    masked_target = mask * position
    masked_current = mask * np.array(end_effector_link.position())

    return (
        np.linalg.norm(masked_current - masked_target) < max_error_pos
        and np.linalg.norm(end_effector_link.world_linear_velocity()) < max_error_vel
    )


def get_unload_position(bucket: scenario_core.Model) -> np.ndarray:

    return bucket.base_position() + np.array([0, 0, 0.3])


class FingersAction(enum.Enum):

    OPEN = enum.auto()
    CLOSE = enum.auto()


def move_fingers(
    panda: gym_ignition_environments.models.panda.Panda, action: FingersAction
) -> None:

    # Get the joints of the fingers
    finger1 = panda.get_joint(joint_name="panda_finger_joint1")
    finger2 = panda.get_joint(joint_name="panda_finger_joint2")

    if action is FingersAction.OPEN:
        finger1.set_position_target(position=finger1.position_limit().max)
        finger2.set_position_target(position=finger2.position_limit().max)

    if action is FingersAction.CLOSE:
        finger1.set_position_target(position=finger1.position_limit().min)
        finger2.set_position_target(position=finger2.position_limit().min)


# ====================
# INITIALIZE THE WORLD
# ====================

# Get the simulator and the world
gazebo, world = gym_ignition.utils.scenario.init_gazebo_sim(
    step_size=0.001, real_time_factor=2.0, steps_per_run=1
)

# Open the GUI
gazebo.gui()
time.sleep(3)
gazebo.run(paused=True)

# Insert the Panda manipulator
panda = gym_ignition_environments.models.panda.Panda(
    world=world, position=[-0.1, 0, 1.0]
)

# Disable joint velocity limits
_ = [j.to_gazebo().set_velocity_limit(1_000) for j in panda.joints()]

# Enable contacts only for the finger links
panda.get_link("panda_leftfinger").to_gazebo().enable_contact_detection(True)
panda.get_link("panda_rightfinger").to_gazebo().enable_contact_detection(True)

# Process model insertion in the simulation
gazebo.run(paused=True)

# Monkey patch the class with finger helpers
panda.open_fingers = partial(move_fingers, panda=panda, action=FingersAction.OPEN)
panda.close_fingers = partial(move_fingers, panda=panda, action=FingersAction.CLOSE)

# Add a custom joint controller to the panda
add_panda_controller(panda=panda, controller_period=gazebo.step_size())

# Populate the world
table = insert_table(world=world)
bucket = insert_bucket(world=world)
gazebo.run(paused=True)

# Create and configure IK for the panda
ik_joints = [
    j.name() for j in panda.joints() if j.type is not scenario_core.JointType_fixed
]
ik = get_panda_ik(panda=panda, optimized_joints=ik_joints)

# Get some manipulator links
finger_left = panda.get_link(link_name="panda_leftfinger")
finger_right = panda.get_link(link_name="panda_rightfinger")
end_effector_frame = panda.get_link(link_name="end_effector_frame")

while True:

    # Insert a new cube
    cube = insert_cube_in_operating_area(world=world)
    gazebo.run(paused=True)

    # =========================
    # PHASE 1: Go over the cube
    # =========================

    print("Hovering")

    # Position over the cube
    position_over_cube = np.array(cube.base_position()) + np.array([0, 0, 0.4])

    # Get the joint configuration that brings the EE over the cube
    over_joint_configuration = solve_ik(
        target_position=position_over_cube,
        target_orientation=np.array(cube.base_orientation()),
        ik=ik,
    )

    # Set the joint references
    assert panda.set_joint_position_targets(over_joint_configuration, ik_joints)

    # Open the fingers
    panda.open_fingers()

    # Run the simulation until the EE reached the desired position
    while not end_effector_reached(
        position=position_over_cube,
        end_effector_link=end_effector_frame,
        max_error_pos=0.05,
        max_error_vel=0.5,
    ):
        gazebo.run()

    # Wait a bit more
    [gazebo.run() for _ in range(500)]

    # =======================
    # PHASE 2: Reach the cube
    # =======================

    print("Reaching")

    # Get the joint configuration that brings the EE to the cube
    over_joint_configuration = solve_ik(
        target_position=np.array(cube.base_position()) + np.array([0, 0, 0.04]),
        target_orientation=np.array(cube.base_orientation()),
        ik=ik,
    )

    # Set the joint references
    assert panda.set_joint_position_targets(over_joint_configuration, ik_joints)
    panda.open_fingers()

    # Run the simulation until the EE reached the desired position
    while not end_effector_reached(
        position=np.array(cube.base_position()) + np.array([0, 0, 0.04]),
        end_effector_link=end_effector_frame,
    ):

        gazebo.run()

    # Wait a bit more
    [gazebo.run() for _ in range(500)]

    # =======================
    # PHASE 3: Grasp the cube
    # =======================

    print("Grasping")

    # Close the fingers
    panda.close_fingers()

    # Detect a graps reading the contact wrenches of the finger links
    while not (
        np.linalg.norm(finger_left.contact_wrench()) >= 50.0
        and np.linalg.norm(finger_right.contact_wrench()) >= 50.0
    ):
        gazebo.run()

    # =============
    # PHASE 4: Lift
    # =============

    print("Lifting")

    # Position over the cube
    position_over_cube = np.array(cube.base_position()) + np.array([0, 0, 0.4])

    # Get the joint configuration that brings the EE over the cube
    over_joint_configuration = solve_ik(
        target_position=position_over_cube,
        target_orientation=np.array(cube.base_orientation()),
        ik=ik,
    )

    # Set the joint references
    assert panda.set_joint_position_targets(over_joint_configuration, ik_joints)

    # Run the simulation until the EE reached the desired position
    while not end_effector_reached(
        position=position_over_cube,
        end_effector_link=end_effector_frame,
        max_error_pos=0.1,
        max_error_vel=0.5,
    ):
        gazebo.run()

    # Wait a bit more
    [gazebo.run() for _ in range(500)]

    # =====================================
    # PHASE 5: Place the cube in the bucket
    # =====================================

    print("Dropping")

    # Get the joint configuration that brings the EE over the bucket
    unload_joint_configuration = solve_ik(
        target_position=get_unload_position(bucket=bucket),
        target_orientation=np.array([0, 1.0, 0, 0]),
        ik=ik,
    )

    # Set the joint references
    assert panda.set_joint_position_targets(unload_joint_configuration, ik_joints)

    # Run the simulation until the EE reached the desired position
    while not end_effector_reached(
        position=get_unload_position(bucket=bucket)
        + np.random.uniform(low=-0.05, high=0.05, size=3),
        end_effector_link=end_effector_frame,
        max_error_pos=0.01,
        max_error_vel=0.1,
        mask=np.array([1, 1, 0]),
    ):

        gazebo.run()

    # Open the fingers
    panda.open_fingers()

    # Wait that both fingers are in not contact (with the cube)
    while finger_left.in_contact() or finger_right.in_contact():
        gazebo.run()

    # Wait a bit more
    [gazebo.run() for _ in range(500)]

    # Remove the cube
    world.remove_model(model_name=cube.name())

# It is always a good practice to close the simulator.
# In this case it is not required since above there is an infinite loop.
# gazebo.close()



# Copyright (C) 2020 Istituto Italiano di Tecnologia (IIT). All rights reserved.
# This software may be modified and distributed under the terms of the
# GNU Lesser General Public License v2.1 or any later version.

import pytest

pytestmark = pytest.mark.gym_ignition

from gym_ignition.utils.math import denormalize, normalize

test_matrix = [
    # Float, None
    (1, None, None, 1),
    (1, 0, None, 1),
    (1, None, 0, 1),
    # Float / Float
    (0, -1, 1, 0),
    (-1, -1, 1, -1),
    (1, -1, 1, 1),
    # List / Float
    ([-1, 0, 1, 2], -2, 2, [-0.5, 0, 0.5, 1]),
    ([-1, 0, 1, 2], -2.0, 2.0, [-0.5, 0, 0.5, 1]),
    ([-1.0, 0, 1, 2], -2, 2, [-0.5, 0, 0.5, 1]),
    ([-1.0, 0, 1, 2], 1, 1, [-1.0, 0, 1, 2]),
    # List / List
    ([-1, 0, 2.0], [-1, -2, 1], [-1, 4, 3], [-1, -0.3333333, 0]),
]


@pytest.mark.parametrize("input,low, high, output", test_matrix)
def test_normalization(input, low, high, output):

    normalized_input = normalize(input=input, low=low, high=high)

    assert output == pytest.approx(normalized_input)
    assert input == pytest.approx(
        denormalize(input=normalized_input, low=low, high=high)
    )



# Copyright (C) 2019 Istituto Italiano di Tecnologia (IIT). All rights reserved.
# This software may be modified and distributed under the terms of the
# GNU Lesser General Public License v2.1 or any later version.

from typing import Dict, List, NewType, Tuple, Union

import gym.spaces
import numpy as np

Done = NewType("Done", bool)
Info = NewType("Info", Dict)
Reward = NewType("Reward", float)
Observation = NewType("Observation", np.ndarray)
Action = NewType("Action", Union[np.ndarray, np.number])

SeedList = NewType("SeedList", List[int])

State = NewType("State", Tuple[Observation, Reward, Done, Info])

ActionSpace = NewType("ActionSpace", gym.spaces.Space)
ObservationSpace = NewType("ObservationSpace", gym.spaces.Space)



# Copyright (C) 2020 Istituto Italiano di Tecnologia (IIT). All rights reserved.
# This software may be modified and distributed under the terms of the
# GNU Lesser General Public License v2.1 or any later version.

import abc
from typing import Union

from gym_ignition import randomizers, utils
from gym_ignition.randomizers import gazebo_env_randomizer
from gym_ignition.randomizers.gazebo_env_randomizer import MakeEnvCallable
from gym_ignition.randomizers.model.sdf import Distribution, Method, UniformParams
from gym_ignition.utils import misc
from gym_ignition_environments import tasks
from gym_ignition_environments.models import cartpole

from scenario import gazebo as scenario

# Tasks that are supported by this randomizer. Used for type hinting.
SupportedTasks = Union[
    tasks.cartpole_discrete_balancing.CartPoleDiscreteBalancing,
    tasks.cartpole_continuous_swingup.CartPoleContinuousSwingup,
    tasks.cartpole_continuous_balancing.CartPoleContinuousBalancing,
]


class CartpoleRandomizersMixin(
    randomizers.abc.TaskRandomizer,
    randomizers.abc.PhysicsRandomizer,
    randomizers.abc.ModelDescriptionRandomizer,
    abc.ABC,
):
    """
    Mixin that collects the implementation of task, model and physics randomizations for
    cartpole environments.
    """

    def __init__(self, randomize_physics_after_rollouts: int = 0):

        # Initialize base classes
        randomizers.abc.TaskRandomizer.__init__(self)
        randomizers.abc.PhysicsRandomizer.__init__(
            self, randomize_after_rollouts_num=randomize_physics_after_rollouts
        )
        randomizers.abc.ModelDescriptionRandomizer.__init__(self)

        # SDF randomizer
        self._sdf_randomizer = None

    # ===========================
    # PhysicsRandomizer interface
    # ===========================

    def get_engine(self):

        return scenario.PhysicsEngine_dart

    def randomize_physics(self, task: SupportedTasks, **kwargs) -> None:

        gravity_z = task.np_random.normal(loc=-9.8, scale=0.2)

        if not task.world.to_gazebo().set_gravity((0, 0, gravity_z)):
            raise RuntimeError("Failed to set the gravity")

    # ========================
    # TaskRandomizer interface
    # ========================

    def randomize_task(self, task: SupportedTasks, **kwargs) -> None:

        # Remove the model from the world
        self._clean_world(task=task)

        if "gazebo" not in kwargs:
            raise ValueError("gazebo kwarg not passed to the task randomizer")

        gazebo = kwargs["gazebo"]

        # Execute a paused run to process model removal
        if not gazebo.run(paused=True):
            raise RuntimeError("Failed to execute a paused Gazebo run")

        # Generate a random model description
        random_model = self.randomize_model_description(task=task)

        # Insert a new model in the world
        self._populate_world(task=task, cartpole_model=random_model)

        # Execute a paused run to process model insertion
        if not gazebo.run(paused=True):
            raise RuntimeError("Failed to execute a paused Gazebo run")

    # ====================================
    # ModelDescriptionRandomizer interface
    # ====================================

    def randomize_model_description(self, task: SupportedTasks, **kwargs) -> str:

        randomizer = self._get_sdf_randomizer(task=task)
        sdf = misc.string_to_file(randomizer.sample())
        return sdf

    # ===============
    # Private Methods
    # ===============

    def _get_sdf_randomizer(
        self, task: SupportedTasks
    ) -> randomizers.model.sdf.SDFRandomizer:

        if self._sdf_randomizer is not None:
            return self._sdf_randomizer

        # Get the model file
        urdf_model_file = cartpole.CartPole.get_model_file()

        # Convert the URDF to SDF
        sdf_model_string = scenario.urdffile_to_sdfstring(urdf_model_file)

        # Write the SDF string to a temp file
        sdf_model = utils.misc.string_to_file(sdf_model_string)

        # Create and initialize the randomizer
        sdf_randomizer = randomizers.model.sdf.SDFRandomizer(sdf_model=sdf_model)

        # Use the RNG of the task
        sdf_randomizer.rng = task.np_random

        # Randomize the mass of all links
        sdf_randomizer.new_randomization().at_xpath("*/link/inertial/mass").method(
            Method.Additive
        ).sampled_from(
            Distribution.Uniform, UniformParams(low=-0.2, high=0.2)
        ).force_positive().add()

        # Process the randomization
        sdf_randomizer.process_data()
        assert len(sdf_randomizer.get_active_randomizations()) > 0

        # Store and return the randomizer
        self._sdf_randomizer = sdf_randomizer
        return self._sdf_randomizer

    @staticmethod
    def _clean_world(task: SupportedTasks) -> None:

        # Remove the model from the simulation
        if task.model_name is not None and task.model_name in task.world.model_names():

            if not task.world.to_gazebo().remove_model(task.model_name):
                raise RuntimeError("Failed to remove the cartpole from the world")

    @staticmethod
    def _populate_world(task: SupportedTasks, cartpole_model: str = None) -> None:

        # Insert a new cartpole.
        # It will create a unique name if there are clashing.
        model = cartpole.CartPole(world=task.world, model_file=cartpole_model)

        # Store the model name in the task
        task.model_name = model.name()


class CartpoleEnvRandomizer(
    gazebo_env_randomizer.GazeboEnvRandomizer, CartpoleRandomizersMixin
):
    """
    Concrete implementation of cartpole environments randomization.
    """

    def __init__(self, env: MakeEnvCallable, num_physics_rollouts: int = 0):

        # Initialize the mixin
        CartpoleRandomizersMixin.__init__(
            self, randomize_physics_after_rollouts=num_physics_rollouts
        )

        # Initialize the environment randomizer
        gazebo_env_randomizer.GazeboEnvRandomizer.__init__(
            self, env=env, physics_randomizer=self
        )



# Copyright (C) 2020 Istituto Italiano di Tecnologia (IIT). All rights reserved.
# This software may be modified and distributed under the terms of the
# GNU Lesser General Public License v2.1 or any later version.

import abc
from typing import Tuple

import gym
import numpy as np
from gym_ignition.base import task
from gym_ignition.utils.typing import (
    Action,
    ActionSpace,
    Observation,
    ObservationSpace,
    Reward,
)

from scenario import core as scenario


class CartPoleContinuousSwingup(task.Task, abc.ABC):
    def __init__(self, agent_rate: float, reward_cart_at_center: bool = True, **kwargs):

        # Initialize the Task base class
        task.Task.__init__(self, agent_rate=agent_rate)

        # Name of the cartpole model
        self.model_name = None

        # Space for resetting the task
        self.reset_space = None

        # Private attributes
        self._reward_cart_at_center = reward_cart_at_center

        # Variables limits
        self._x_threshold = 2.4  # m
        self._dx_threshold = 20.0  # m /s
        self._q_threshold = np.deg2rad(5 * 360)  # rad
        self._dq_threshold = np.deg2rad(3 * 360)  # rad / s

    def create_spaces(self) -> Tuple[ActionSpace, ObservationSpace]:

        # Create the action space
        max_force = 200.0  # Nm
        action_space = gym.spaces.Box(
            low=np.array([-max_force]), high=np.array([max_force]), dtype=np.float32
        )

        # Configure reset limits
        high = np.array(
            [
                self._x_threshold,  # x
                self._dx_threshold,  # dx
                self._q_threshold,  # q
                self._dq_threshold,  # dq
            ]
        )

        # Configure the reset space
        self.reset_space = gym.spaces.Box(low=-high, high=high, dtype=np.float32)

        # Configure the observation space
        obs_high = high.copy() * 1.2
        observation_space = gym.spaces.Box(
            low=-obs_high, high=obs_high, dtype=np.float32
        )

        return action_space, observation_space

    def set_action(self, action: Action) -> None:

        # Get the force value
        force = action.tolist()[0]

        # Set the force value
        model = self.world.get_model(self.model_name)
        ok_force = model.get_joint("linear").set_generalized_force_target(force)

        if not ok_force:
            raise RuntimeError("Failed to set the force to the cart")

    def get_observation(self) -> Observation:

        # Get the model
        model = self.world.get_model(self.model_name)

        # Get the new joint positions and velocities
        q, x = model.joint_positions(["pivot", "linear"])
        dq, dx = model.joint_velocities(["pivot", "linear"])

        # Create the observation
        observation = Observation(np.array([x, dx, q, dq]))

        # Return the observation
        return observation

    def get_reward(self) -> Reward:

        # Get the model
        model = self.world.get_model(self.model_name)

        # Get the pendulum position
        q = model.get_joint("pivot").position()

        # Get the cart state
        x = model.get_joint("linear").position()
        dx = model.get_joint("linear").velocity()

        # Reward is [0, 1] for q=[0, pi]
        reward = (np.cos(q) + 1) / 2

        # Penalize cart velocities
        reward -= 0.1 * (dx ** 2)

        # Penalize positions close to the end of the rail
        reward -= 10.0 * (x >= 0.8 * self._x_threshold)

        return reward

    def is_done(self) -> bool:

        # Get the observation
        observation = self.get_observation()

        # The environment is done if the observation is outside its space
        done = not self.reset_space.contains(observation)

        return done

    def reset_task(self) -> None:

        if self.model_name not in self.world.model_names():
            raise RuntimeError("Cartpole model not found in the world")

        # Get the model
        model = self.world.get_model(self.model_name)

        # Control the cart in force mode
        linear = model.get_joint("linear")
        ok_control_mode = linear.set_control_mode(scenario.JointControlMode_force)

        if not ok_control_mode:
            raise RuntimeError("Failed to change the control mode of the cartpole")

        # Create a new cartpole state
        q = np.pi - np.deg2rad(self.np_random.uniform(low=-60, high=60))
        x, dx, dq = self.np_random.uniform(low=-0.05, high=0.05, size=(3,))

        # Reset the cartpole state
        ok_reset_pos = model.to_gazebo().reset_joint_positions(
            [x, q], ["linear", "pivot"]
        )
        ok_reset_vel = model.to_gazebo().reset_joint_velocities(
            [dx, dq], ["linear", "pivot"]
        )

        if not (ok_reset_pos and ok_reset_vel):
            raise RuntimeError("Failed to reset the cartpole state")



# Copyright (C) 2019 Istituto Italiano di Tecnologia (IIT). All rights reserved.
# This software may be modified and distributed under the terms of the
# GNU Lesser General Public License v2.1 or any later version.

import dataclasses
from typing import Tuple

import gym_ignition_models
import pytest
from gym_ignition.utils import misc

from scenario import gazebo as scenario


def id_gazebo_fn(val: Tuple[float, float, float]):
    if isinstance(val, tuple) and len(val) == 3:
        return f"step={val[0]}_rtf={val[1]}_iters={val[2]}"


@pytest.fixture(scope="function")
def gazebo_fixture(request):
    """
    Return an instance of the GazeboSimulator ensuring that it
    is closed even in case of failure.

    Example:

        from ..common.utils import gazebo_fixture as gazebo

        @pytest.mark.parametrize("gazebo", [(0.001, 1.0, 1)], indirect=True)
        def test_foo(gazebo: scenario.gazebo.GazeboSimulator):

            assert gazebo.initialize()
            # ...

        @pytest.mark.parametrize("gazebo, name",
                                 [((0.001, 1.0, 1), "name_1"),
                                  ((0.001, 1.0, 1), "name_2"),
                                  ((0.001, 1.0, 1), "name_3")],
                                  indirect="gazebo")
        def test_foo(gazebo: scenario.gazebo.GazeboSimulator, name: str):

            model_name = name
            assert gazebo.initialize()
            # ...
    """

    step_size, rtf, iterations = request.param
    gazebo = scenario.GazeboSimulator(step_size, rtf, iterations)

    yield gazebo

    gazebo.close()


@pytest.fixture(scope="function")
def default_world_fixture(request):
    """
    Initialize a default world with ground and physics.

    Example:

        from ..common.utils import default_world_fixture as default_world

        @pytest.mark.parametrize("default_world", [(0.001, 1.0, 1)], indirect=True)
        def test_foo(default_world: Tuple[scenario.GazeboSimulator, scenario.World]):

            # Get the simulator and the world
            gazebo, world = default_world

            # ...
    """

    step_size, rtf, iterations = request.param
    gazebo = scenario.GazeboSimulator(step_size, rtf, iterations)

    assert gazebo.initialize()

    world = gazebo.get_world().to_gazebo()
    assert world.insert_model(gym_ignition_models.get_model_file("ground_plane"))
    assert world.set_physics_engine(scenario.PhysicsEngine_dart)

    yield gazebo, world

    gazebo.close()


def get_multi_world_sdf_file() -> str:

    multi_world_sdf_string = f"""
    <?xml version="1.0" ?>
    <sdf version="1.6">
        <world name="world1">
        </world>
        <world name="world2">
        </world>
    </sdf>"""

    multi_world_sdf = misc.string_to_file(multi_world_sdf_string)
    return multi_world_sdf


def get_cube_sdf_string() -> str:
    return """
    <?xml version="1.0" ?>
    <sdf version="1.6">
        <model name='box'>
        <pose>0 0 0.5 0 -0 0</pose>
            <link name='box_link'>
                <inertial>
                <inertia>
                    <ixx>1</ixx>
                    <ixy>0</ixy>
                    <ixz>0</ixz>
                    <iyy>1</iyy>
                    <iyz>0</iyz>
                    <izz>1</izz>
                </inertia>
                <mass>1</mass>
                </inertial>
                <collision name='box_collision'>
                <geometry>
                    <box>
                        <size>1 1 1</size>
                    </box>
                </geometry>
                <surface>
                    <friction>
                        <ode/>
                    </friction>
                    <contact/>
                </surface>
                </collision>
                <visual name='box_visual'>
                <geometry>
                    <box>
                    <size>1 1 1</size>
                    </box>
                </geometry>
                <material>
                    <ambient>1 0 0 1</ambient>
                    <diffuse>1 0 0 1</diffuse>
                    <specular>1 0 0 1</specular>
                </material>
                </visual>
            </link>
        </model>
    </sdf>
    """


def get_cube_urdf_string() -> str:

    mass = 5.0
    edge = 0.2
    i = 1 / 12 * mass * (edge ** 2 + edge ** 2)
    cube_urdf = f"""
    <robot name="cube_robot" xmlns:xacro="http://www.ros.org/wiki/xacro">
        <link name="cube">
            <inertial>
              <origin rpy="0 0 0" xyz="0 0 0"/>
              <mass value="{mass}"/>
              <inertia ixx="{i}" ixy="0" ixz="0" iyy="{i}" iyz="0" izz="{i}"/>
            </inertial>
            <visual>
              <geometry>
                <box size="{edge} {edge} {edge}"/>
              </geometry>
              <origin rpy="0 0 0" xyz="0 0 0"/>
            </visual>
            <collision>
              <geometry>
                <box size="{edge} {edge} {edge}"/>
              </geometry>
              <origin rpy="0 0 0" xyz="0 0 0"/>
            </collision>
        </link>
    </robot>"""

    return cube_urdf


def get_cube_urdf() -> str:

    model_file = misc.string_to_file(get_cube_urdf_string())
    return model_file


def get_empty_world_sdf() -> str:

    world_sdf_string = scenario.get_empty_world()

    world_file = misc.string_to_file(world_sdf_string)
    return world_file


@dataclasses.dataclass
class SphereURDF:

    mass: float = 5.0
    radius: float = 0.1
    restitution: float = 0

    def urdf(self) -> str:

        i = 2.0 / 5 * self.mass * self.radius * self.radius
        urdf = f"""
            <robot name="sphere_model" xmlns:xacro="http://www.ros.org/wiki/xacro">
                <link name="sphere">
                    <inertial>
                      <origin rpy="0 0 0" xyz="0 0 0"/>
                      <mass value="{self.mass}"/>
                      <inertia ixx="{i}" ixy="0" ixz="0" iyy="{i}" iyz="0" izz="{i}"/>
                    </inertial>
                    <visual>
                      <geometry>
                        <sphere radius="{self.radius}"/>
                      </geometry>
                      <origin rpy="0 0 0" xyz="0 0 0"/>
                    </visual>
                    <collision>
                      <geometry>
                        <sphere radius="{self.radius}"/>
                      </geometry>
                      <origin rpy="0 0 0" xyz="0 0 0"/>
                    </collision>
                </link>
                <gazebo reference="sphere">
                  <collision><surface><bounce>
                    <restitution_coefficient>{self.restitution}</restitution_coefficient>
                  </bounce></surface></collision>
                </gazebo>
            </robot>
            """

        return urdf



# Copyright (C) 2019 Istituto Italiano di Tecnologia (IIT). All rights reserved.
# This software may be modified and distributed under the terms of the
# GNU Lesser General Public License v2.1 or any later version.

import gym
import numpy as np
from gym_ignition import gympp_bindings as bindings
from gym_ignition.utils import logger, resource_finder


def test_joint_controller():
    # ==========
    # PARAMETERS
    # ==========

    agent_rate = 100.0
    physics_rate = 1000.0
    controller_rate = 500.0

    plugin_data = bindings.PluginData()
    plugin_data.libName = "RobotController"
    plugin_data.className = "scenario::plugins::gazebo::RobotController"

    # Find and load the model SDF file
    model_sdf_file = resource_finder.find_resource("CartPole/CartPole.urdf")
    with open(model_sdf_file, "r") as stream:
        model_sdf_string = stream.read()

    # Initialize the model data
    model_data = bindings.ModelInitData()
    model_data.sdfString = model_sdf_string

    # Get the model name
    model_name = bindings.GazeboSimulator.getModelNameFromSDF(model_sdf_string)
    robot_name = model_name

    # =============
    # CONFIGURATION
    # =============

    # Create the gazebo wrapper
    num_of_iterations = int(physics_rate / agent_rate)
    desired_rtf = float(np.finfo(np.float32).max)
    gazebo = bindings.GazeboSimulator(num_of_iterations, desired_rtf, physics_rate)
    assert gazebo, "Failed to get the gazebo wrapper"

    # Set the verbosity
    logger.set_level(gym.logger.DEBUG)

    # Initialize the world
    world_ok = gazebo.setupGazeboWorld("DefaultEmptyWorld.world")
    assert world_ok, "Failed to initialize the gazebo world"

    # Initialize the ignition gazebo wrapper (creates a paused simulation)
    gazebo_initialized = gazebo.initialize()
    assert gazebo_initialized, "Failed to initialize ignition gazebo"

    # Insert the model
    model_ok = gazebo.insertModel(model_data, plugin_data)
    assert model_ok, "Failed to insert the model in the simulation"

    assert bindings.RobotSingleton_get().exists(
        robot_name
    ), "The robot interface was not registered in the singleton"

    # Extract the robot interface from the singleton
    robot_weak_ptr = bindings.RobotSingleton_get().getRobot(robot_name)
    assert not robot_weak_ptr.expired(), "The Robot object has expired"
    assert (
        robot_weak_ptr.lock()
    ), "The returned Robot object does not contain a valid interface"
    assert robot_weak_ptr.lock().valid(), "The Robot object is not valid"

    # Get the pointer to the robot interface
    robot = robot_weak_ptr.lock()

    # Set the default update rate
    robot.setdt(1 / controller_rate)

    # Control the cart joint in position
    ok_cm = robot.setJointControlMode("linear", bindings.JointControlMode_Position)
    assert ok_cm, "Failed to control the cart joint in position"

    # Set the PID of the cart joint
    pid = bindings.PID(10000, 1000, 1000)
    pid_ok = robot.setJointPID("linear", pid)
    assert pid_ok, "Failed to set the PID of the cart joint"

    # Reset the robot state
    robot.resetJoint("pivot", 0, 0)
    robot.resetJoint("linear", 0, 0)

    # Generate the cart trajectory
    cart_ref = np.fromiter(
        (0.2 * np.sin(2 * np.pi * 0.5 * t) for t in np.arange(0, 5, 1 / agent_rate)),
        dtype=np.float,
    )

    # Initialize the cart position buffer
    pos_cart_buffer = np.zeros(np.shape(cart_ref))

    for (i, ref) in enumerate(cart_ref):
        # Set the references
        ok1 = robot.setJointPositionTarget("linear", ref)
        assert ok1, "Failed to set joint references"

        # Step the simulator
        gazebo.run()

        # Read the position
        pos_cart_buffer[i] = robot.jointPosition("linear")

    # Check that the trajectory was followed correctly
    assert (
        np.abs(pos_cart_buffer - cart_ref).sum() / cart_ref.size < 5e-3
    ), "The reference trajectory was not tracked correctly"



# Copyright (C) 2020 Istituto Italiano di Tecnologia (IIT). All rights reserved.
# This software may be modified and distributed under the terms of the
# GNU Lesser General Public License v2.1 or any later version.

import pytest

pytestmark = pytest.mark.scenario

import gym_ignition_models
import numpy as np
from gym_ignition.utils import misc

from scenario import core as scenario_core
from scenario import gazebo as scenario_gazebo

from ..common import utils
from ..common.utils import gazebo_fixture as gazebo

# Set the verbosity
scenario_gazebo.set_verbosity(scenario_gazebo.Verbosity_debug)


@pytest.mark.parametrize(
    "gazebo", [(0.001, 1.0, 1)], indirect=True, ids=utils.id_gazebo_fn
)
def test_insert_multiple_worlds(gazebo: scenario_gazebo.GazeboSimulator):

    empty_world_sdf = utils.get_empty_world_sdf()
    assert gazebo.insert_world_from_sdf(empty_world_sdf, "myWorld1")

    assert not gazebo.insert_world_from_sdf(empty_world_sdf, "myWorld1")
    assert gazebo.insert_world_from_sdf(empty_world_sdf, "myWorld2")

    assert gazebo.initialize()
    assert "myWorld1" in gazebo.world_names()
    assert "myWorld2" in gazebo.world_names()

    my_world1 = gazebo.get_world("myWorld1")
    my_world2 = gazebo.get_world("myWorld2")

    assert my_world1.id() != my_world2.id()
    assert my_world1.name() == "myWorld1"
    assert my_world2.name() == "myWorld2"


@pytest.mark.parametrize(
    "gazebo", [(0.001, 1.0, 1)], indirect=True, ids=utils.id_gazebo_fn
)
def test_insert_multiple_world(gazebo: scenario_gazebo.GazeboSimulator):

    multi_world_sdf = utils.get_multi_world_sdf_file()

    assert gazebo.insert_worlds_from_sdf(multi_world_sdf)
    assert gazebo.initialize()

    assert "world1" in gazebo.world_names()
    assert "world2" in gazebo.world_names()

    world1 = gazebo.get_world("world1")
    world2 = gazebo.get_world("world2")

    assert world1.name() == "world1"
    assert world2.name() == "world2"

    assert world1.id() != world2.id()


@pytest.mark.parametrize(
    "gazebo", [(0.001, 1.0, 1)], indirect=True, ids=utils.id_gazebo_fn
)
def test_insert_multiple_world_rename(gazebo: scenario_gazebo.GazeboSimulator):

    multi_world_sdf = utils.get_multi_world_sdf_file()

    assert not gazebo.insert_worlds_from_sdf(multi_world_sdf, ["only_one_name"])
    assert gazebo.insert_worlds_from_sdf(multi_world_sdf, ["myWorld1", "myWorld2"])
    assert gazebo.initialize()

    assert "myWorld1" in gazebo.world_names()
    assert "myWorld2" in gazebo.world_names()

    world1 = gazebo.get_world("myWorld1")
    world2 = gazebo.get_world("myWorld2")

    assert world1.name() == "myWorld1"
    assert world2.name() == "myWorld2"

    assert world1.id() != world2.id()


@pytest.mark.parametrize(
    "gazebo", [(0.001, 1.0, 1)], indirect=True, ids=utils.id_gazebo_fn
)
def test_insert_world_multiple_calls(gazebo: scenario_gazebo.GazeboSimulator):

    single_world_sdf = utils.get_empty_world_sdf()

    assert gazebo.insert_world_from_sdf(single_world_sdf)
    assert not gazebo.insert_world_from_sdf(single_world_sdf)
    assert gazebo.insert_world_from_sdf(single_world_sdf, "world2")
    assert gazebo.initialize()

    assert "default" in gazebo.world_names()
    assert "world2" in gazebo.world_names()

    world1 = gazebo.get_world("default")
    world2 = gazebo.get_world("world2")

    assert world1.name() == "default"
    assert world2.name() == "world2"

    assert world1.id() != world2.id()


# # This test is flaky
# @pytest.mark.xfail(strict=False)
# @pytest.mark.parametrize("gazebo, solver",
#                          [((0.001, 2.0, 1), "pgs"),
#                           ((0.001, 2.0, 1), "dantzig")],
#                          indirect=["gazebo"],
#                          ids=utils.id_gazebo_fn)
# def test_multi_world_simulation(gazebo: scenario_gazebo.GazeboSimulator,
#                                 solver: str):
#
#     # Empty DART world with bullet as collision detector.
#     # It should prevent ODE crashes in a multi-world setting due to its static variables.
#     world_sdf_string = f"""
#     <?xml version="1.0" ?>
#     <sdf version="1.7">
#         <world name="default">
#             <physics default="true" type="dart">
#                 <dart>
#                     <collision_detector>bullet</collision_detector>
#                     <solver>
#                         <solver_type>{solver}</solver_type>
#                     </solver>
#                 </dart>
#             </physics>
#             <light type="directional" name="sun">
#                 <cast_shadows>true</cast_shadows>
#                 <pose>0 0 10 0 0 0</pose>
#                 <diffuse>1 1 1 1</diffuse>
#                 <specular>0.5 0.5 0.5 1</specular>
#                 <attenuation>
#                     <range>1000</range>
#                     <constant>0.9</constant>
#                     <linear>0.01</linear>
#                     <quadratic>0.001</quadratic>
#                 </attenuation>
#                 <direction>-0.5 0.1 -0.9</direction>
#             </light>
#         </world>
#     </sdf>
#     """
#
#     # Create a tmp file from the SDF string
#     world_sdf_file = misc.string_to_file(string=world_sdf_string)
#
#     # Load two different worlds
#     assert gazebo.insert_world_from_sdf(world_sdf_file, "dart1")
#     assert gazebo.insert_world_from_sdf(world_sdf_file, "dart2")
#
#     # Initialize the simulator
#     assert gazebo.initialize()
#
#     # gazebo.gui()
#     # import time
#     # time.sleep(1)
#     # gazebo.run(paused=True)
#
#     sphere_urdf_string = utils.SphereURDF(restitution=0.8).urdf()
#     sphere_urdf = misc.string_to_file(sphere_urdf_string)
#     ground_plane_urdf = gym_ignition_models.get_model_file(robot_name="ground_plane")
#
#     # Pose of the sphere
#     sphere_pose = scenario_core.Pose([0, 0, 0.5], [1, 0, 0, 0])
#
#     # Populate the first DART world
#     world1 = gazebo.get_world("dart1")
#     world1.set_physics_engine(scenario_gazebo.PhysicsEngine_dart)
#     world1.insert_model(ground_plane_urdf)
#     world1.insert_model(sphere_urdf, sphere_pose)
#     sphere1 = world1.get_model(model_name="sphere_model")
#
#     # Populate the second DART world
#     world2 = gazebo.get_world("dart2")
#     world2.set_physics_engine(scenario_gazebo.PhysicsEngine_dart)
#     world2.insert_model(ground_plane_urdf)
#     world2.insert_model(sphere_urdf, sphere_pose)
#     sphere2 = world2.get_model(model_name="sphere_model")
#
#     # Integration time
#     dt = gazebo.step_size() * gazebo.steps_per_run()
#
#     # Enable contacts
#     sphere1.enable_contacts(True)
#     sphere2.enable_contacts(True)
#
#     for _ in np.arange(start=0.0, stop=5.0, step=dt):
#
#         # Run the simulator
#         assert gazebo.run()
#
#         # Check that both worlds evolve similarly
#         assert sphere1.base_position() == \
#                pytest.approx(sphere2.base_position(), abs=0.001)
#         assert sphere1.base_world_linear_velocity() == \
#                pytest.approx(sphere2.base_world_linear_velocity(), abs=0.001)
#         assert sphere1.base_world_angular_velocity() == \
#                pytest.approx(sphere2.base_world_angular_velocity())



# Copyright (C) 2019 Istituto Italiano di Tecnologia (IIT). All rights reserved.
# This software may be modified and distributed under the terms of the
# GNU Lesser General Public License v2.1 or any later version.

import numpy as np
import pytest

from . import utils


@pytest.mark.parametrize("simulator_name", ["gazebo", "pybullet"])
def test_robot_fixed_base(simulator_name: str):

    if simulator_name == "gazebo":
        simulator = utils.Gazebo(physics_rate=1000)
    elif simulator_name == "pybullet":
        simulator = utils.PyBullet(physics_rate=1000)
    else:
        raise ValueError(simulator_name)

    # Get the pendulum robot and specify its base position
    base_position = np.array([-3, 3, 2])
    robot = utils.get_pendulum(simulator=simulator, base_position=base_position)

    # Robot should not fall due to gravity
    for _ in range(100):
        simulator.step()

    assert np.allclose(robot.base_pose()[0], base_position, atol=1e-3)
    assert np.allclose(robot.base_pose()[1], np.array([1.0, 0, 0, 0]), atol=1e-3)


# @pytest.mark.parametrize("simulator_name", ["gazebo", "pybullet"])
# def test_robot_floating_base(simulator_name: str):
#
#     if simulator_name == "gazebo":
#         simulator = utils.Gazebo(physics_rate=1000)
#     elif simulator_name == "pybullet":
#         simulator = utils.PyBullet(physics_rate=1000)
#     else:
#         raise ValueError(simulator_name)
#
#     # Get the pendulum robot and specify its base position
#     base_position = np.array([-3, 3, 2])
#     robot = utils.get_pendulum(simulator=simulator, base_position=base_position)
#
#     assert np.allclose(robot.base_pose()[0], base_position, atol=1e-3)
#     assert np.allclose(robot.base_pose()[1], np.array([1.0, 0, 0, 0]), atol=1e-3)
#
#     old_pos_z = base_position[2]
#
#     # Robot will fall due to gravity
#     for _ in range(100):
#         simulator.step()
#         current_pos_z = robot.base_pose()[0][2]
#         assert current_pos_z < old_pos_z
#         old_pos_z = current_pos_z


@pytest.mark.parametrize("simulator_name", ["pybullet"])
def test_robot_floating_to_fixed(simulator_name: str):

    if simulator_name == "gazebo":
        simulator = utils.Gazebo(physics_rate=1000)
    elif simulator_name == "pybullet":
        simulator = utils.PyBullet(physics_rate=1000)
    else:
        raise ValueError(simulator_name)

    # Get the pendulum robot and specify its base position
    base_position = np.array([-3, 3, 2])
    robot = utils.get_pendulum(simulator=simulator, base_position=base_position)

    # Let's make the robot fall a little
    for _ in range(10):
        simulator.step()

    current_base_position, current_base_orientation = robot.base_pose()
    assert current_base_position[2] < base_position[2]

    ok_fixed_base = robot.set_as_floating_base(False)
    assert ok_fixed_base, "Failed to set the robot as fixed base"

    ok_reset_base_pose = robot.reset_base_pose(
        current_base_position, current_base_orientation
    )
    assert ok_reset_base_pose, "Failed to reset the base pose"

    # Robot should now not fall due to gravity
    for _ in range(100):
        simulator.step()

    assert np.allclose(robot.base_pose()[0], current_base_position, atol=1e-3)
    assert np.allclose(robot.base_pose()[1], current_base_orientation, atol=1e-3)



# Copyright (C) 2020 Istituto Italiano di Tecnologia (IIT). All rights reserved.
# This software may be modified and distributed under the terms of the
# GNU Lesser General Public License v2.1 or any later version.

import os
from dataclasses import dataclass
from enum import Enum, auto
from typing import Dict, List, Optional, Tuple, Union

import idyntree.bindings as idt
import numpy as np
from gym_ignition import rbd


class TargetType(Enum):

    POSITION = auto()
    ROTATION = auto()
    POSE = auto()


@dataclass
class TransformTargetData:

    position: np.ndarray
    quaternion: np.ndarray


@dataclass
class TargetData:

    type: TargetType
    weight: Union[float, Tuple[float, float]]
    data: Union[np.ndarray, TransformTargetData]


@dataclass
class IKSolution:

    joint_configuration: np.ndarray
    base_position: np.ndarray = np.array([0.0, 0.0, 0.0])
    base_quaternion: np.ndarray = np.array([1.0, 0.0, 0.0, 0.0])


class RotationParametrization(Enum):

    QUATERNION = auto()
    ROLL_PITCH_YAW = auto()

    def to_idyntree(self):

        if self.value == RotationParametrization.QUATERNION.value:
            return idt.InverseKinematicsRotationParametrizationQuaternion

        elif self.value == RotationParametrization.ROLL_PITCH_YAW.value:
            return idt.InverseKinematicsRotationParametrizationRollPitchYaw

        else:
            raise ValueError(self.value)


class TargetResolutionMode(Enum):

    TARGET_AS_CONSTRAINT_FULL = auto()
    TARGET_AS_CONSTRAINT_NONE = auto()
    TARGET_AS_CONSTRAINT_POSITION = auto()
    TARGET_AS_CONSTRAINT_ROTATION = auto()

    def to_idyntree(self):

        if self.value == TargetResolutionMode.TARGET_AS_CONSTRAINT_FULL.value:
            return idt.InverseKinematicsTreatTargetAsConstraintFull

        elif self.value == TargetResolutionMode.TARGET_AS_CONSTRAINT_NONE.value:
            return idt.InverseKinematicsTreatTargetAsConstraintNone

        elif self.value == TargetResolutionMode.TARGET_AS_CONSTRAINT_POSITION.value:
            return idt.InverseKinematicsTreatTargetAsConstraintPositionOnly

        elif self.value == TargetResolutionMode.TARGET_AS_CONSTRAINT_ROTATION.value:
            return idt.InverseKinematicsTreatTargetAsConstraintRotationOnly

        else:
            raise ValueError(self.value)


class InverseKinematicsNLP:
    def __init__(
        self,
        urdf_filename: str,
        considered_joints: List[str] = None,
        joint_serialization: List[str] = None,
    ) -> None:

        self._floating_base: bool = False
        self._base_frame: Optional[str] = None
        self._urdf_filename: str = urdf_filename
        self._targets_data: Dict[str, TargetData] = dict()
        self._ik: Optional[idt.InverseKinematics] = None

        self._considered_joints: List[str] = considered_joints
        self._joint_serialization: List[str] = joint_serialization

    # ======================
    # INITIALIZATION METHODS
    # ======================

    def initialize(
        self,
        rotation_parametrization: RotationParametrization = RotationParametrization.ROLL_PITCH_YAW,
        target_mode: TargetResolutionMode = TargetResolutionMode.TARGET_AS_CONSTRAINT_NONE,
        cost_tolerance: float = 1e-08,
        constraints_tolerance: float = 1e-4,
        max_iterations: int = 1000,
        base_frame: str = None,
        floating_base: bool = False,
        verbosity: int = 1,
    ) -> None:

        # Create the IK object
        self._ik = idt.InverseKinematics()

        # Load the URDF model and get the model loader object.
        # We create the full model with all the joints specified in joint_serialization.
        model_loader: idt.ModelLoader = self._get_model_loader(
            urdf=self._urdf_filename, joint_serialization=self._joint_serialization
        )

        # Get the model
        model: idt.Model = model_loader.model()

        # If all joints are enabled, get the list of joint names in order to know the
        # serialization of the full IK solution
        if self._joint_serialization is None:
            self._joint_serialization = []

            for joint_idx in range(model.getNrOfJoints()):
                self._joint_serialization.append(model.getJointName(joint_idx))

        # Configure the considered joints for the optimization.
        # If not specified, use the serialization of the full solution.
        if self._considered_joints is None:
            self._considered_joints = self._joint_serialization

        # Set the model in the IK specifying the considered joints
        if not self._ik.setModel(model, self._considered_joints):
            raise RuntimeError("Failed to set the model in the IK object")

        # Configure IK
        self._ik.setVerbosity(verbosity)
        self._ik.setMaxIterations(max_iterations)
        self._ik.setCostTolerance(cost_tolerance)
        self._ik.setConstraintsTolerance(constraints_tolerance)
        self._ik.setDefaultTargetResolutionMode(target_mode.to_idyntree())
        self._ik.setRotationParametrization(rotation_parametrization.to_idyntree())

        # Optionally change the base frame
        if base_frame is not None:
            # Store the frame of the base link
            self._base_frame = base_frame

            # Change the base frame
            if not self._ik.setFloatingBaseOnFrameNamed(base_frame):
                raise RuntimeError("Failed to change floating base frame")
        else:
            self._base_frame = model.getLinkName(model.getDefaultBaseLink())

        # Store whether the IK optimize a fixed or floating base robot
        self._floating_base = floating_base

        if not self._floating_base:

            # Add a frame constraint for the base
            self.add_frame_transform_constraint(
                frame_name=self._base_frame,
                position=np.array([0.0, 0, 0.0]),
                quaternion=np.array([1.0, 0, 0, 0]),
            )

    def set_current_robot_configuration(
        self,
        base_position: np.ndarray,
        base_quaternion: np.ndarray,
        joint_configuration: np.ndarray,
    ) -> None:

        if joint_configuration.size != len(self._joint_serialization):
            raise ValueError(joint_configuration)

        H = rbd.idyntree.numpy.FromNumPy.to_idyntree_transform(
            position=base_position, quaternion=base_quaternion
        )

        q = rbd.idyntree.numpy.FromNumPy.to_idyntree_dyn_vector(
            array=joint_configuration
        )

        if not self._ik.setCurrentRobotConfiguration(H, q):
            raise RuntimeError("Failed to set the current robot configuration")

        if not self._floating_base:

            self.update_frame_transform_constraint(
                frame_name=self._base_frame,
                position=base_position,
                quaternion=base_quaternion,
            )

    def set_current_joint_configuration(
        self, joint_name: str, configuration: float
    ) -> None:

        if joint_name not in self._joint_serialization:
            raise ValueError(joint_name)

        if not self._ik.setJointConfiguration(
            jointName=joint_name, jointConfiguration=configuration
        ):
            raise RuntimeError(
                f"Failed to set the configuration of joint '{joint_name}'"
            )

    # ==============
    # TARGET METHODS
    # ==============

    def add_target(
        self,
        frame_name: str,
        target_type: TargetType,
        weight: Union[float, Tuple[float, float]] = None,
        as_constraint: bool = False,
    ) -> None:

        # Check the type of the 'weight' argument
        float_target_types = {TargetType.ROTATION, TargetType.POSITION}
        weight_type = float if target_type in float_target_types else tuple

        # Backward compatibility: if the target type is POSE and the weight is a float,
        # we apply the same weight to both target components
        if target_type is TargetType.POSE and isinstance(weight, float):
            weight = (weight, weight)

        # Set the default weight if not specified
        default_weight = 1.0 if target_type in float_target_types else (1.0, 1.0)
        weight = weight if weight is not None else default_weight

        if not isinstance(weight, weight_type):
            raise ValueError(f"The weight must be {weight_type} for this target")

        if target_type == TargetType.ROTATION:
            # Add the target
            ok_target = self._ik.addRotationTarget(
                frame_name, idt.Rotation.Identity(), weight
            )

            # Initialize the target data buffers
            self._targets_data[frame_name] = TargetData(
                type=TargetType.ROTATION, weight=weight, data=np.array([1.0, 0, 0, 0])
            )

        elif target_type == TargetType.POSITION:
            # Add the target
            ok_target = self._ik.addPositionTarget(
                frame_name, idt.Position_Zero(), weight
            )

            # Initialize the target data buffers
            self._targets_data[frame_name] = TargetData(
                type=TargetType.POSITION, weight=weight, data=np.array([0.0, 0, 0])
            )

        elif target_type == TargetType.POSE:
            # Add the target
            ok_target = self._ik.addTarget(
                frame_name, idt.Transform.Identity(), weight[0], weight[1]
            )

            # Create the transform target data
            target_data = TransformTargetData(
                position=np.array([0.0, 0, 0]), quaternion=np.array([1.0, 0, 0, 0])
            )

            # Initialize the target data buffers
            self._targets_data[frame_name] = TargetData(
                type=TargetType.POSE, weight=weight, data=target_data
            )

        else:
            raise ValueError(target_type)

        if not ok_target:
            raise RuntimeError(f"Failed to add target for frame '{frame_name}'")

        if as_constraint:
            if target_type == TargetType.ROTATION:
                constraint = idt.InverseKinematicsTreatTargetAsConstraintRotationOnly
            elif target_type == TargetType.POSITION:
                constraint = idt.InverseKinematicsTreatTargetAsConstraintPositionOnly
            else:
                assert target_type == TargetType.POSE
                constraint = idt.InverseKinematicsTreatTargetAsConstraintFull

            if not self._ik.setTargetResolutionMode(frame_name, constraint):
                raise RuntimeError(f"Failed to set target '{frame_name}' as constraint")

    def add_com_target(
        self,
        weight: float = 1.0,
        as_constraint: bool = False,
        constraint_tolerance: float = 1e-8,
    ) -> None:

        # Add the target
        self._ik.setCOMTarget(idt.Position_Zero(), weight)

        # Configure it either as target or constraint
        self._ik.setCOMAsConstraint(asConstraint=as_constraint)
        self._ik.setCOMAsConstraintTolerance(tolerance=constraint_tolerance)

        # Initialize the target data buffers
        assert "com" not in self._targets_data.keys()
        self._targets_data["com"] = TargetData(
            type=TargetType.POSITION, weight=weight, data=np.array([0.0, 0, 0])
        )

    def update_position_target(self, target_name: str, position: np.ndarray) -> None:

        if target_name not in self.get_active_target_names(
            target_type=TargetType.POSITION
        ):

            raise ValueError(f"Failed to find a position target '{target_name}'")

        # Create the iDynTree position
        p = rbd.idyntree.numpy.FromNumPy.to_idyntree_position(position=position)

        # Update the target inside IK
        if not self._ik.updateTarget(target_name, p):
            raise RuntimeError(f"Failed to update position of target '{target_name}'")

        # Get the configured weight
        weight = self._targets_data[target_name].weight

        # Update the target data
        self._targets_data[target_name] = TargetData(
            type=TargetType.POSITION, weight=weight, data=position
        )

    def update_rotation_target(self, target_name: str, quaternion: np.ndarray) -> None:

        if target_name not in self.get_active_target_names(
            target_type=TargetType.ROTATION
        ):

            raise ValueError(f"Failed to find a rotation target '{target_name}'")

        # Create the iDynTree rotation matrix
        R = rbd.idyntree.numpy.FromNumPy.to_idyntree_rotation(quaternion=quaternion)

        # Update the target inside IK
        if not self._ik.updateRotationTarget(target_name, R):
            raise RuntimeError(f"Failed to update rotation of target '{target_name}'")

        # Get the configured weight
        weight = self._targets_data[target_name].weight

        # Update the target data
        self._targets_data[target_name] = TargetData(
            type=TargetType.ROTATION, weight=weight, data=quaternion
        )

    def update_transform_target(
        self, target_name: str, position: np.ndarray, quaternion: np.ndarray
    ) -> None:

        if target_name not in self.get_active_target_names(target_type=TargetType.POSE):

            raise ValueError(f"Failed to find a transform target '{target_name}'")

        # Create the iDynTree transform
        H = rbd.idyntree.numpy.FromNumPy.to_idyntree_transform(
            position=position, quaternion=quaternion
        )

        # Update the target inside IK
        if not self._ik.updateTarget(target_name, H):
            raise RuntimeError(f"Failed to update transform of target '{target_name}'")

        # Get the configured weight
        weight = self._targets_data[target_name].weight

        # Create the transform target data
        transform_data = TransformTargetData(position=position, quaternion=quaternion)

        # Update the target data
        self._targets_data[target_name] = TargetData(
            type=TargetType.POSE, weight=weight, data=transform_data
        )

    def update_com_target(self, position: np.ndarray) -> None:

        if not self._ik.isCOMTargetActive():
            raise RuntimeError("Constraint on CoM not active")

        # Create the iDynTree position
        p = rbd.idyntree.numpy.FromNumPy.to_idyntree_position(position=position)

        # Update the target inside IK
        self._ik.setCOMTarget(p, self._targets_data["com"].weight)

        # Update the target data
        self._targets_data["com"] = TargetData(
            type=TargetType.POSITION,
            weight=self._targets_data["com"].weight,
            data=position,
        )

    def deactivate_com_target(self) -> None:

        if not self._ik.isCOMTargetActive():
            raise RuntimeError("Constraint on CoM not active")

        self._ik.deactivateCOMTarget()

    # =============
    # FRAME METHODS
    # =============

    def add_frame_transform_constraint(
        self, frame_name: str, position: np.ndarray, quaternion: np.ndarray
    ) -> None:

        # Create the transform
        H = rbd.idyntree.numpy.FromNumPy.to_idyntree_transform(
            position=position, quaternion=quaternion
        )

        # Add the target
        if not self._ik.addFrameConstraint(frame_name, H):
            raise RuntimeError(f"Failed to add constraint on frame '{frame_name}'")

    def add_frame_position_constraint(
        self, frame_name: str, position: np.ndarray
    ) -> None:

        # Create the position
        p = rbd.idyntree.numpy.FromNumPy.to_idyntree_position(position=position)

        # Add the constraint
        if not self._ik.addFramePositionConstraint(frame_name, p):
            raise RuntimeError(f"Failed to add constraint on frame '{frame_name}'")

    def add_frame_rotation_constraint(
        self, frame_name: str, quaternion: np.ndarray
    ) -> None:

        # Create the position
        R = rbd.idyntree.numpy.FromNumPy.to_idyntree_rotation(quaternion=quaternion)

        # Add the constraint
        if not self._ik.addFrameRotationConstraint(frame_name, R):
            raise RuntimeError(f"Failed to add constraint on frame '{frame_name}'")

    def update_frame_transform_constraint(
        self, frame_name: str, position: np.ndarray, quaternion: np.ndarray
    ) -> None:

        if not self._ik.isFrameConstraintActive(frame_name):
            raise RuntimeError(f"Constraint on frame '{frame_name}' not active")

        # Create the transform
        H = rbd.idyntree.numpy.FromNumPy.to_idyntree_transform(
            position=position, quaternion=quaternion
        )

        if not self._ik.activateFrameConstraint(frame_name, H):
            raise RuntimeError(f"Failed to update constraint on frame '{frame_name}'")

    def deactivate_frame_constraint(self, frame_name: str) -> None:

        if not self._ik.isFrameConstraintActive(frame_name):
            raise RuntimeError(f"Constraint on frame '{frame_name}' not active")

        if not self._ik.deactivateFrameConstraint(frame_name):
            raise RuntimeError(
                f"Failed to deactivate constraint on frame '{frame_name}'"
            )

    # ===================
    # IK SOLUTION METHODS
    # ===================

    def solve(self) -> None:

        if not self._ik.solve():
            raise RuntimeError("Failed to solve IK")

        # Initialize next solver call
        self._warm_start_with_last_solution()

    def warm_start_from(
        self, full_solution: IKSolution = None, reduced_solution: IKSolution = None
    ) -> None:

        if (
            full_solution is None
            and reduced_solution is None
            or full_solution is not None
            and reduced_solution is not None
        ):
            raise ValueError("You have to specify either a full or a reduced solution")

        if (
            reduced_solution is not None
            and reduced_solution.joint_configuration.size
            != len(self._considered_joints)
        ):
            raise RuntimeError(
                "The joint configuration does not match the number of considered joints"
            )

        if full_solution is not None and full_solution.joint_configuration.size != len(
            self._joint_serialization
        ):
            raise RuntimeError(
                "The joint configuration does not match the number of model joints"
            )

        solution = reduced_solution if reduced_solution is not None else full_solution

        H = rbd.idyntree.numpy.FromNumPy.to_idyntree_transform(
            position=solution.base_position, quaternion=solution.base_quaternion
        )

        q = rbd.idyntree.numpy.FromNumPy.to_idyntree_dyn_vector(
            array=solution.joint_configuration
        )

        # Warm start the solver
        if not self._ik.setFullJointsInitialCondition(H, q):
            raise RuntimeError("Failed to warm start the IK solver")

    # =======
    # GETTERS
    # =======

    def get_base_frame(self) -> str:

        return self._base_frame

    def get_available_target_names(self) -> List[str]:

        # Get the reduced model
        model = self._ik.reducedModel()

        # Note that also frames (modeled as fake links) are available targets
        link_names = []

        for link_index in range(model.getNrOfLinks()):
            link_names.append(model.getLinkName(link_index))

        return link_names

    def get_active_target_names(self, target_type: TargetType = None) -> List[str]:

        if target_type is None:
            return list(self._targets_data.keys())

        else:
            return [
                name
                for name, value in self._targets_data.items()
                if value.type == target_type
            ]

    def get_target_data(self, target_name: str) -> TargetData:

        return self._targets_data[target_name]

    def get_full_solution(self) -> IKSolution:

        # Initialize buffers
        base_transform = idt.Transform.Identity()
        joint_positions = idt.VectorDynSize(self._ik.fullModel().getNrOfJoints())
        assert len(joint_positions) == len(self._joint_serialization)

        # Get the solution
        self._ik.getFullJointsSolution(base_transform, joint_positions)

        # Convert to numpy objects
        joint_positions = joint_positions.toNumPy()
        base_position = base_transform.getPosition().toNumPy()
        base_quaternion = base_transform.getRotation().asQuaternion().toNumPy()

        return IKSolution(
            base_position=base_position,
            base_quaternion=base_quaternion,
            joint_configuration=joint_positions,
        )

    def get_reduced_solution(self) -> IKSolution:

        # Initialize buffers
        base_transform = idt.Transform.Identity()
        joint_positions = idt.VectorDynSize(self._ik.reducedModel().getNrOfJoints())
        assert len(joint_positions) == len(self._considered_joints)

        # Get the solution
        self._ik.getReducedSolution(base_transform, joint_positions)

        # Convert to numpy objects
        joint_positions = joint_positions.toNumPy()
        base_position = base_transform.getPosition().toNumPy()
        base_quaternion = base_transform.getRotation().asQuaternion().toNumPy()

        return IKSolution(
            base_position=base_position,
            base_quaternion=base_quaternion,
            joint_configuration=joint_positions,
        )

    # ===============
    # PRIVATE METHODS
    # ===============

    @staticmethod
    def _get_model_loader(
        urdf: str, joint_serialization: List[str] = None
    ) -> idt.ModelLoader:

        # Get the model loader
        model_loader = idt.ModelLoader()

        if not os.path.exists(urdf):
            raise FileNotFoundError(urdf)

        # Load the model
        if joint_serialization:
            ok_load = model_loader.loadReducedModelFromFile(urdf, joint_serialization)
        else:
            ok_load = model_loader.loadModelFromFile(urdf)

        if not ok_load:
            raise RuntimeError("Failed to load model")

        # Due to some SWIG internal, returning the model contained by the loader
        # does not work as expected
        return model_loader

    def _warm_start_with_last_solution(self) -> None:

        last_solution = self.get_full_solution()
        self.warm_start_from(full_solution=last_solution)



import abc
from typing import Tuple

import numpy as np
from scipy.spatial.transform import Rotation


class Transform(abc.ABC):
    @staticmethod
    def from_position_and_quaternion(
        position: np.ndarray, quaternion: np.ndarray
    ) -> np.ndarray:

        if quaternion.size != 4:
            raise ValueError("Quaternion array must have 4 elements")

        rotation = Quaternion.to_rotation(quaternion)
        transform = Transform.from_position_and_rotation(
            position=position, rotation=rotation
        )

        return transform

    @staticmethod
    def from_position_and_rotation(
        position: np.ndarray, rotation: np.ndarray
    ) -> np.ndarray:

        if position.size != 3:
            raise ValueError("Position array must have 3 elements")

        if rotation.shape != (3, 3):
            raise ValueError("Rotation must be a square 3x3 matrix")

        transform = np.eye(4)
        transform[0:3, 3] = position
        transform[0:3, 0:3] = rotation

        return transform

    @staticmethod
    def to_position_and_rotation(
        transform: np.ndarray,
    ) -> Tuple[np.ndarray, np.ndarray]:

        if transform.shape != (4, 4):
            raise ValueError("Transform must be a 4x4 matrix")

        position = transform[0:3, 3]
        rotation = transform[0:3, 0:3]

        return position, rotation

    @staticmethod
    def to_position_and_quaternion(
        transform: np.ndarray,
    ) -> Tuple[np.ndarray, np.ndarray]:

        p, R = Transform.to_position_and_rotation(transform=transform)
        return p, Quaternion.from_matrix(matrix=R)


class Quaternion(abc.ABC):
    @staticmethod
    def to_wxyz(xyzw: np.ndarray) -> np.ndarray:

        if xyzw.shape != (4,):
            raise ValueError(xyzw)

        return xyzw[[3, 0, 1, 2]]

    @staticmethod
    def to_xyzw(wxyz: np.ndarray) -> np.ndarray:

        if wxyz.shape != (4,):
            raise ValueError(wxyz)

        return wxyz[[1, 2, 3, 0]]

    @staticmethod
    def to_rotation(quaternion: np.ndarray) -> np.ndarray:

        if quaternion.shape != (4,):
            raise ValueError(quaternion)

        xyzw = Quaternion.to_xyzw(quaternion)

        return Rotation.from_quat(xyzw).as_matrix()

    @staticmethod
    def from_matrix(matrix: np.ndarray) -> np.ndarray:

        if matrix.shape != (3, 3):
            raise ValueError(matrix)

        quaternion_xyzw = Rotation.from_matrix(matrix).as_quat()
        quaternion_wxyz = Quaternion.to_wxyz(quaternion_xyzw)

        return quaternion_wxyz



# Copyright (C) 2019 Istituto Italiano di Tecnologia (IIT). All rights reserved.
# This software may be modified and distributed under the terms of the
# GNU Lesser General Public License v2.1 or any later version.

from . import (
    cartpole_continuous_balancing,
    cartpole_continuous_swingup,
    cartpole_discrete_balancing,
    pendulum_swingup,
)



# Copyright (C) 2020 Istituto Italiano di Tecnologia (IIT). All rights reserved.
# This software may be modified and distributed under the terms of the
# GNU Lesser General Public License v2.1 or any later version.

from typing import List

from gym_ignition.scenario import model_with_file, model_wrapper
from gym_ignition.utils.scenario import get_unique_model_name

from scenario import core as scenario


class Pendulum(model_wrapper.ModelWrapper, model_with_file.ModelWithFile):
    def __init__(
        self,
        world: scenario.World,
        position: List[float] = (0.0, 0.0, 0.0),
        orientation: List[float] = (1.0, 0, 0, 0),
        model_file: str = None,
    ):

        # Get a unique model name
        model_name = get_unique_model_name(world, "pendulum")

        # Initial pose
        initial_pose = scenario.Pose(position, orientation)

        # Get the model file (URDF or SDF) and allow to override it
        if model_file is None:
            model_file = Pendulum.get_model_file()

        # Insert the model
        ok_model = world.to_gazebo().insert_model(model_file, initial_pose, model_name)

        if not ok_model:
            raise RuntimeError("Failed to insert model")

        # Get the model
        model = world.get_model(model_name)

        # Initialize base class
        super().__init__(model=model)

    @classmethod
    def get_model_file(cls) -> str:

        import gym_ignition_models

        return gym_ignition_models.get_model_file("pendulum")



# Copyright (C) 2021 Istituto Italiano di Tecnologia (IIT). All rights reserved.
# This software may be modified and distributed under the terms of the
# GNU Lesser General Public License v2.1 or any later version.

import numpy as np


def wedge(vector3: np.ndarray) -> np.ndarray:
    """
    Convert a 3D vector to a skew-symmetric matrix.

    Args:
        vector3: The 3D vector defining the matrix coefficients.

    Returns:
        The skew-symmetric matrix whose elements are created from the input vector.

    Note:
        The wedge operator can be useful to compute the cross product of 3D vectors:
        :math:`v_1 \\times v_2 = v_1^\\wedge v_2`.
    """

    if vector3.size != 3:
        raise ValueError(vector3)

    s = np.zeros(shape=(3, 3))
    s[1, 0] = vector3[2]
    s[0, 1] = -vector3[2]
    s[0, 2] = vector3[1]
    s[2, 0] = -vector3[1]
    s[2, 1] = vector3[0]
    s[1, 2] = -vector3[0]

    return s


def vee(matrix3x3: np.ndarray) -> np.ndarray:
    """
    Convert a 3x3 matrix to a 3D vector with the components of its skew-symmetric part.

    Args:
        matrix3x3: The input 3x3 matrix. If present, its symmetric part is removed.

    Returns:
        The 3D vector defining the skew-symmetric matrix.

    Note:
        This is the inverse operator of :py:func:`wedge`.
    """

    if matrix3x3.shape != (3, 3):
        raise ValueError(matrix3x3)

    skew_symmetric = extract_skew(matrix3x3)

    return np.array([skew_symmetric[2, 1], skew_symmetric[0, 2], skew_symmetric[1, 0]])


def extract_skew(matrix: np.ndarray) -> np.ndarray:
    """
    Extract the skew-symmetric part of a square matrix.

    Args:
        matrix: A square matrix.

    Returns:
        The skew-symmetric part of the input matrix.
    """

    if matrix.shape[0] != matrix.shape[1]:
        raise ValueError(matrix)

    return 0.5 * (matrix - matrix.T)


def extract_symm(matrix: np.ndarray) -> np.ndarray:
    """
    Extract the symmetric part of a square matrix.

    Args:
        matrix: A square matrix.

    Returns:
        The symmetric part of the input matrix.
    """

    if matrix.shape[0] != matrix.shape[1]:
        raise ValueError(matrix)

    return 0.5 * (matrix + matrix.T)



# Copyright (C) 2020 Istituto Italiano di Tecnologia (IIT). All rights reserved.
# This software may be modified and distributed under the terms of the
# GNU Lesser General Public License v2.1 or any later version.

import numpy
from gym.envs.registration import register

from . import models, randomizers, tasks

max_float = float(numpy.finfo(numpy.float32).max)

register(
    id="Pendulum-Gazebo-v0",
    entry_point="gym_ignition.runtimes.gazebo_runtime:GazeboRuntime",
    max_episode_steps=5000,
    kwargs={
        "task_cls": tasks.pendulum_swingup.PendulumSwingUp,
        "agent_rate": 1000,
        "physics_rate": 1000,
        "real_time_factor": max_float,
    },
)

register(
    id="CartPoleDiscreteBalancing-Gazebo-v0",
    entry_point="gym_ignition.runtimes.gazebo_runtime:GazeboRuntime",
    max_episode_steps=5000,
    kwargs={
        "task_cls": tasks.cartpole_discrete_balancing.CartPoleDiscreteBalancing,
        "agent_rate": 1000,
        "physics_rate": 1000,
        "real_time_factor": max_float,
    },
)

register(
    id="CartPoleContinuousBalancing-Gazebo-v0",
    entry_point="gym_ignition.runtimes.gazebo_runtime:GazeboRuntime",
    max_episode_steps=5000,
    kwargs={
        "task_cls": tasks.cartpole_continuous_balancing.CartPoleContinuousBalancing,
        "agent_rate": 1000,
        "physics_rate": 1000,
        "real_time_factor": max_float,
    },
)

register(
    id="CartPoleContinuousSwingup-Gazebo-v0",
    entry_point="gym_ignition.runtimes.gazebo_runtime:GazeboRuntime",
    max_episode_steps=5000,
    kwargs={
        "task_cls": tasks.cartpole_continuous_swingup.CartPoleContinuousSwingup,
        "agent_rate": 1000,
        "physics_rate": 1000,
        "real_time_factor": max_float,
    },
)



# Copyright (C) 2019 Istituto Italiano di Tecnologia (IIT). All rights reserved.
# This software may be modified and distributed under the terms of the
# GNU Lesser General Public License v2.1 or any later version.

import tempfile


def string_to_file(string: str) -> str:

    handle, tmpfile = tempfile.mkstemp()

    with open(handle, "w") as f:
        f.write(string)

    return tmpfile



# Copyright (C) 2020 Istituto Italiano di Tecnologia (IIT). All rights reserved.
# This software may be modified and distributed under the terms of the
# GNU Lesser General Public License v2.1 or any later version.

import abc
from typing import List

from gym_ignition import scenario
from gym_ignition.utils.scenario import get_unique_model_name

from scenario import core as scenario_core


class ICubGazeboABC(
    scenario.model_wrapper.ModelWrapper, scenario.model_with_file.ModelWithFile, abc.ABC
):

    DOFS = 32
    NUM_LINKS = 39
    NUM_JOINTS = 32

    initial_positions = {
        # Left leg
        "l_knee": -1.05,
        "l_ankle_pitch": -0.57,
        "l_ankle_roll": -0.024,
        "l_hip_pitch": 0.48,
        "l_hip_roll": 0.023,
        "l_hip_yaw": -0.005,
        # Left arm
        "l_elbow": 0.54,
        "l_wrist_pitch": 0.0,
        "l_wrist_prosup": 0.0,
        "l_wrist_yaw": 0.0,
        "l_shoulder_pitch": -0.159,
        "l_shoulder_roll": 0.435,
        "l_shoulder_yaw": 0.183,
        # Head
        "neck_pitch": 0.0,
        "neck_roll": 0.0,
        "neck_yaw": 0.0,
        # Right leg
        "r_knee": -1.05,
        "r_ankle_pitch": -0.57,
        "r_ankle_roll": -0.024,
        "r_hip_pitch": 0.48,
        "r_hip_roll": 0.023,
        "r_hip_yaw": -0.005,
        # Right arm
        "r_elbow": 0.54,
        "r_wrist_pitch": 0.0,
        "r_wrist_prosup": 0.0,
        "r_wrist_yaw": 0.0,
        "r_shoulder_pitch": -0.159,
        "r_shoulder_roll": 0.435,
        "r_shoulder_yaw": 0.183,
        # Torso
        "torso_pitch": 0.1,
        "torso_roll": 0.0,
        "torso_yaw": 0.0,
    }

    def __init__(
        self,
        world: scenario_core.World,
        position: List[float],
        orientation: List[float],
        model_file: str = None,
    ):

        # Get a unique model name
        model_name = get_unique_model_name(world, "icub")

        # Initial pose
        initial_pose = scenario_core.Pose(position, orientation)

        # Get the model file (URDF or SDF) and allow to override it
        if model_file is None:
            model_file = self.get_model_file()

        # Insert the model
        ok_model = world.to_gazebo().insert_model(model_file, initial_pose, model_name)

        if not ok_model:
            raise RuntimeError("Failed to insert model")

        # Get the model
        model = world.get_model(model_name)

        # Initialize base class
        super().__init__(model=model)

        # Store the initial positions
        q0 = list(self.initial_positions.values())
        joint_names = list(self.initial_positions.keys())
        assert self.dofs() == len(q0) == len(joint_names)

        ok_q0 = self.to_gazebo().reset_joint_positions(q0, joint_names)
        assert ok_q0, "Failed to set initial position"


class ICubGazebo(ICubGazeboABC):
    def __init__(
        self,
        world: scenario_core.World,
        position: List[float] = (0.0, 0.0, 0.572),
        orientation: List[float] = (0, 0, 0, 1.0),
        model_file: str = None,
    ):

        super().__init__(
            world=world,
            position=position,
            orientation=orientation,
            model_file=model_file,
        )

    @classmethod
    def get_model_file(cls) -> str:

        import gym_ignition_models

        return gym_ignition_models.get_model_file("iCubGazeboV2_5")


class ICubGazeboSimpleCollisions(ICubGazeboABC):
    def __init__(
        self,
        world: scenario_core.World,
        position: List[float] = (0.0, 0.0, 0.572),
        orientation: List[float] = (0, 0, 0, 1.0),
        model_file: str = None,
    ):

        super().__init__(
            world=world,
            position=position,
            orientation=orientation,
            model_file=model_file,
        )

    @classmethod
    def get_model_file(cls) -> str:

        import gym_ignition_models

        return gym_ignition_models.get_model_file("iCubGazeboSimpleCollisionsV2_5")



# Copyright (C) 2020 Istituto Italiano di Tecnologia (IIT). All rights reserved.
# This software may be modified and distributed under the terms of the
# GNU Lesser General Public License v2.1 or any later version.

from dataclasses import dataclass, field
from typing import Iterable, List, Tuple

from gym_ignition.context.gazebo import plugin

GRAVITY = (0, 0, -9.80665)


@dataclass
class ComputedTorqueFixedBase(plugin.GazeboPlugin):

    urdf: str
    kp: List[float]
    ki: List[float]
    kd: List[float]
    joints: List[str]
    gravity: Tuple[float, float, float] = field(default_factory=lambda: GRAVITY)

    # Private fields
    _name: str = field(init=False, repr=False, default="ComputedTorqueFixedBase")
    _plugin_name: str = field(init=False, repr=False, default="ControllerRunner")
    _plugin_class: str = field(
        init=False, repr=False, default="scenario::plugins::gazebo::ControllerRunner"
    )

    def to_xml(self) -> str:
        xml = f"""
        <controller name="{self._name}">
            <kp>{self._to_str(self.kp)}</kp>
            <ki>{self._to_str(self.ki)}</ki>
            <kd>{self._to_str(self.kd)}</kd>
            <urdf>{self.urdf}</urdf>
            <joints>{self._to_str(self.joints)}</joints>
            <gravity>{self._to_str(self.gravity)}</gravity>
        </controller>
        """

        return xml

    @staticmethod
    def _to_str(iterable: Iterable) -> str:

        return " ".join([str(el) for el in iterable])



# Copyright (C) 2020 Istituto Italiano di Tecnologia (IIT). All rights reserved.
# This software may be modified and distributed under the terms of the
# GNU Lesser General Public License v2.1 or any later version.

from gym_ignition.base.robot.robot_joints import JointControlMode

from . import utils


def test_joint_force():
    # Get the simulator
    gazebo = utils.Gazebo(physics_rate=1000, iterations=1)

    # Insert the robot
    pendulum = utils.get_pendulum(simulator=gazebo)

    # Configure the robot
    ok_mode = pendulum.set_joint_control_mode("pivot", JointControlMode.TORQUE)
    assert ok_mode

    # Step the simulator
    gazebo.step()

    # No references set
    assert pendulum.joint_force("pivot") == 0.0

    # Step the simulator
    gazebo.step()

    # Set a reference. No force reference yet applied.
    torque = 42.42
    pendulum.set_joint_force("pivot", torque)
    assert pendulum.joint_force("pivot") == 0.0

    # Step the simulator. Now the force reference should have been actuated and after
    # the physics step it should be returned by the method.
    gazebo.step()
    assert pendulum.joint_force("pivot") == torque

    # Step again the simulator. No force reference has been specified and the method
    # should return zero.
    gazebo.step()
    assert pendulum.joint_force("pivot") == 0.0

    # Close the simulator
    gazebo.close()


def test_joint_force_multiple_iterations():
    # Get the simulator
    gazebo = utils.Gazebo(physics_rate=1000, iterations=2)

    # Insert the robot
    pendulum = utils.get_pendulum(simulator=gazebo)
    pendulum.set_joint_control_mode("pivot", JointControlMode.TORQUE)

    # Step the simulator
    gazebo.step()

    # No references set
    assert pendulum.joint_force("pivot") == 0.0

    # Step the simulator
    gazebo.step()

    # Set a reference. No force reference yet applied.
    torque = 42.42
    pendulum.set_joint_force("pivot", torque)
    assert pendulum.joint_force("pivot") == 0.0

    # Step the simulator.
    # Note that since gazebo was configured with multiple iterations,
    # the force is applied only in the first one and we still read zero!
    gazebo.step()
    assert pendulum.joint_force("pivot") == 0.0

    # Step again the simulator
    gazebo.step()
    assert pendulum.joint_force("pivot") == 0.0

    # Close the simulator
    gazebo.close()



# Copyright (C) 2020 Istituto Italiano di Tecnologia (IIT). All rights reserved.
# This software may be modified and distributed under the terms of the
# GNU Lesser General Public License v2.1 or any later version.

import pytest

pytestmark = pytest.mark.scenario

from typing import Tuple

import gym_ignition_models
import numpy as np

from scenario import core as scenario_core
from scenario import gazebo as scenario_gazebo

from ..common.utils import default_world_fixture as default_world

# Set the verbosity
scenario_gazebo.set_verbosity(scenario_gazebo.Verbosity_debug)


@pytest.mark.parametrize("default_world", [(1.0 / 1_000, 1.0, 1)], indirect=True)
def test_velocity_direct(
    default_world: Tuple[scenario_gazebo.GazeboSimulator, scenario_gazebo.World]
):

    # Get the simulator and the world
    gazebo, world = default_world

    # Get the URDF model
    pendulum_urdf = gym_ignition_models.get_model_file("pendulum")

    # Insert a pendulum model
    assert world.insert_model(pendulum_urdf)
    assert "pendulum" in world.model_names()

    # Get the model and cast it to Gazebo
    pendulum = world.get_model("pendulum").to_gazebo()

    # Add some friction
    assert pendulum.get_joint("pivot").to_gazebo().set_coulomb_friction(value=0.01)
    assert pendulum.get_joint("pivot").to_gazebo().set_viscous_friction(value=0.2)

    # Show the GUI
    # import time
    # gazebo.gui()
    # time.sleep(2)

    # Get the pivot joint
    assert "pivot" in pendulum.joint_names()
    pivot = pendulum.get_joint(joint_name="pivot")

    # Reset the joint to 90 degs and make it swing
    assert pivot.to_gazebo().reset_position(position=np.deg2rad(90))
    [gazebo.run() for _ in range(5_000)]

    # It should rest in its stable equilibrium point
    assert np.deg2rad(179.7) <= pivot.position() <= np.deg2rad(180.3)

    # Control the joint in velocity direct
    assert pivot.set_control_mode(
        mode=scenario_core.JointControlMode_velocity_follower_dart
    )

    # Set the velocity reference to 3.14 rad / s
    assert pivot.set_velocity_target(velocity=np.pi)

    # The target should be reached after just one run
    assert gazebo.run()
    assert pivot.velocity() == pytest.approx(np.pi)

    # Simulate for a while
    [gazebo.run() for _ in range(1_500)]

    # Check again
    assert pivot.velocity() == pytest.approx(np.pi)

    # Change direction abruptly.
    # Due to existing joint velocity limits, it takes two steps to reach the target.
    # Note: even by increasing the limit to its maximum value, two steps are necessary.
    assert pivot.set_velocity_target(velocity=-np.pi)
    assert gazebo.run()
    assert gazebo.run()
    assert pivot.velocity() == pytest.approx(-np.pi)

    # Go in idle and make it swing
    assert pivot.set_control_mode(mode=scenario_core.JointControlMode_idle)
    [gazebo.run() for _ in range(5_000)]

    # It should rest in its stable equilibrium point
    assert (
        2 * np.pi + np.deg2rad(179.7)
        <= pivot.position()
        <= 2 * np.pi + np.deg2rad(180.3)
    )



# Copyright (C) 2020 Istituto Italiano di Tecnologia (IIT). All rights reserved.
# This software may be modified and distributed under the terms of the
# GNU Lesser General Public License v2.1 or any later version.

import abc
from typing import Callable, Dict, Optional, Union, cast

import gym
from gym_ignition.randomizers.abc import PhysicsRandomizer, TaskRandomizer
from gym_ignition.randomizers.physics import dart
from gym_ignition.runtimes import gazebo_runtime
from gym_ignition.utils import typing

MakeEnvCallable = Callable[[Optional[Dict]], gym.Env]


class GazeboEnvRandomizer(gym.Wrapper, TaskRandomizer, abc.ABC):
    """
    Base class to implement an environment randomizer for Ignition Gazebo.

    The randomizer is a :py:class:`gym.Wrapper` that extends the
    :py:meth:`gym.Env.reset` method. Objects that inherit from this class are used to
    setup the environment for the handled :py:class:`~gym_ignition.base.task.Task`.

    In its simplest form, a randomizer populates the world with all the models that need
    to be part of the simulation. The task could then operate on them from a
    :py:class:`~scenario.core.Model` object.

    More complex environments may require to randomize one or more simulated entities.
    Concrete classes that implement a randomizer could use
    :py:class:`~gym_ignition.randomizers.model.sdf.SDFRandomizer` to randomize the model
    and :py:class:`~gym_ignition.randomizers.abc.PhysicsRandomizer` to randomize
    the physics.

    Args:
        env: Defines the environment to handle. This argument could be either the string
            id if the environment does not need to be registered or a function that
            returns an environment object.
        physics_randomizer: Object that randomizes physics. The default physics engine is
            DART with no randomizations.

    Note:
        In order to randomize physics, the handled
        :py:class:`scenario.gazebo.GazeboSimulator` is destroyed and created again.
        This operation is demanding, consider randomizing physics at a low rate.
    """

    def __init__(
        self,
        env: Union[str, MakeEnvCallable],
        physics_randomizer: PhysicsRandomizer = dart.DART(),
        **kwargs,
    ):

        # Print the extra kwargs
        gym.logger.debug(f"GazeboEnvRandomizer: {dict(kwargs=kwargs)}")

        # Store the options
        self._env_option = env
        self._kwargs = dict(**kwargs, physics_engine=physics_randomizer.get_engine())

        # Create the environment
        env_to_wrap = self._create_environment(env=self._env_option, **self._kwargs)

        # Initialize the wrapper
        gym.Wrapper.__init__(self, env=env_to_wrap)

        # Store the physics randomizer
        self._physics_randomizer = physics_randomizer

    # ===============
    # gym.Env methods
    # ===============

    def reset(self, **kwargs) -> typing.Observation:

        # Reset the physics
        if self._physics_randomizer.physics_expired():

            # Get the random components of the task
            seed = self.env.task.seed
            np_random = self.env.task.np_random

            # Reset the runtime + task, creating a new Gazebo instance
            self.env.close()
            del self.env
            self.env = self._create_environment(self._env_option, **self._kwargs)

            # Restore the random components
            self.env.seed(seed=seed)
            assert self.env.task.seed == seed
            self.env.task.np_random = np_random

        # Mark the beginning of a new rollout
        self._physics_randomizer.increase_rollout_counter()

        # Reset the task through the TaskRandomizer
        self.randomize_task(task=self.env.task, gazebo=self.env.gazebo, **kwargs)

        ok_paused_run = self.env.gazebo.run(paused=True)

        if not ok_paused_run:
            raise RuntimeError("Failed to execute a paused Gazebo run")

        # Reset the Task
        return self.env.reset()

    # ===============
    # Private methods
    # ===============

    def _create_environment(
        self, env: Union[str, MakeEnvCallable], **kwargs
    ) -> gazebo_runtime.GazeboRuntime:

        if isinstance(env, str):
            env_to_wrap = self._create_from_id(env_id=env, **kwargs)

        elif callable(env):
            env_to_wrap = self._create_from_callable(make_env=env, **kwargs)

        else:
            raise ValueError("The type of env object was not recognized")

        if not isinstance(env_to_wrap.unwrapped, gazebo_runtime.GazeboRuntime):
            raise ValueError("The environment to wrap is not a GazeboRuntime")

        return cast(gazebo_runtime.GazeboRuntime, env_to_wrap)

    @staticmethod
    def _create_from_callable(make_env: MakeEnvCallable, **kwargs) -> gym.Env:

        env = make_env(**kwargs)
        return env

    @staticmethod
    def _create_from_id(env_id: str, **kwargs) -> gym.Env:

        env = gym.make(env_id, **kwargs)
        return env



# Copyright (C) 2020 Istituto Italiano di Tecnologia (IIT). All rights reserved.
# This software may be modified and distributed under the terms of the
# GNU Lesser General Public License v2.1 or any later version.

import pytest

pytestmark = pytest.mark.scenario

import gym_ignition_models

from scenario import gazebo as scenario

from ..common import utils
from ..common.utils import gazebo_fixture as gazebo

# Set the verbosity
scenario.set_verbosity(scenario.Verbosity_debug)


@pytest.mark.parametrize(
    "gazebo",
    [
        (0.001, 1.0, 1),
        (0.1, 5.0, 5),
        (0.001, 0.0, 1),
        (0.001, -1.0, 1),
        (0.001, 1.0, 0),
        (0, 1.0, 1),
    ],
    indirect=True,
    ids=utils.id_gazebo_fn,
)
def test_initialization(gazebo: scenario.GazeboSimulator):

    ok = gazebo.initialize()

    rtf = gazebo.real_time_factor()
    step_size = gazebo.step_size()
    iterations = gazebo.steps_per_run()

    if step_size <= 0:
        assert not ok
        assert not gazebo.initialized()
        assert not gazebo.run()

    if rtf <= 0:
        assert not ok
        assert not gazebo.initialized()
        assert not gazebo.run()

    if iterations <= 0:
        assert not ok
        assert not gazebo.initialized()
        assert not gazebo.run()

    if rtf > 0 and iterations > 0 and step_size > 0:
        assert ok
        assert gazebo.initialized()


@pytest.mark.parametrize(
    "gazebo", [(0.001, 1.0, 1)], indirect=True, ids=utils.id_gazebo_fn
)
def test_run(gazebo: scenario.GazeboSimulator):

    assert gazebo.initialize()
    assert gazebo.run(paused=True)
    assert gazebo.run()


@pytest.mark.parametrize(
    "gazebo", [(0.001, 1.0, 1)], indirect=True, ids=utils.id_gazebo_fn
)
def test_pause(gazebo: scenario.GazeboSimulator):

    assert gazebo.initialize()
    assert not gazebo.running()
    assert gazebo.pause()

    assert gazebo.initialize()

    assert not gazebo.running()
    assert gazebo.pause()

    assert gazebo.run(paused=True)
    assert not gazebo.running()
    assert gazebo.pause()

    assert gazebo.run(paused=False)
    assert not gazebo.running()
    assert gazebo.pause()


@pytest.mark.parametrize(
    "gazebo", [(0.001, 1.0, 1)], indirect=True, ids=utils.id_gazebo_fn
)
def test_paused_step(gazebo: scenario.GazeboSimulator):

    assert gazebo.initialize()

    world = gazebo.get_world().to_gazebo()
    assert world.insert_model(gym_ignition_models.get_model_file("ground_plane"))
    assert "ground_plane" in world.model_names()
    gazebo.run(paused=True)

    world.remove_model("ground_plane")

    assert "ground_plane" in world.model_names()
    gazebo.run(paused=True)
    assert "ground_plane" not in world.model_names()
    assert world.time() == 0.0


@pytest.mark.parametrize(
    "gazebo", [(0.001, 1.0, 1)], indirect=True, ids=utils.id_gazebo_fn
)
def test_load_default_world(gazebo: scenario.GazeboSimulator):

    assert gazebo.initialize()
    assert gazebo.world_names()
    assert len(gazebo.world_names()) == 1

    world1 = gazebo.get_world()
    assert world1
    assert world1.name() in gazebo.world_names()

    world2 = gazebo.get_world(gazebo.world_names()[0])
    assert world2

    assert world1.id() == world2.id()
    assert world1.name() == world2.name()

    # TODO: understand how to compare shared ptr returned by swig with nullptr
    # world3 = gazebo.get_world("foo")
    # assert world3



# Copyright (C) 2020 Istituto Italiano di Tecnologia (IIT). All rights reserved.
# This software may be modified and distributed under the terms of the
# GNU Lesser General Public License v2.1 or any later version.

from typing import List, Tuple, Union

import gym.spaces
import gym_ignition_models
import numpy as np

from scenario import core
from scenario import gazebo as scenario


def get_unique_model_name(world: scenario.World, model_name: str) -> str:
    """
    Get a unique model name given a world configuration.

    This function find the first available model name starting from the argument and
    appending a integer postfix until the resulting name is unique in the world.

    Tentatives example: `cartpole`, `cartpole1`, `cartpole2`, ...

    Args:
        world: An initialized world.
        model_name: The first model name attempt.

    Raises:
        ValueError: If the world is not valid.

    Returns:
        The unique model name calculated from the original name.
    """

    if world.id() == 0:
        raise ValueError("The world is not valid")

    postfix = 0
    model_name_tentative = f"{model_name}"

    while model_name_tentative in world.model_names():

        postfix += 1
        model_name_tentative = f"{model_name}{postfix}"

    return model_name_tentative


def init_gazebo_sim(
    step_size: float = 0.001, real_time_factor: float = 1.0, steps_per_run: int = 1
) -> Tuple[scenario.GazeboSimulator, Union[scenario.World, core.World]]:
    """
    Initialize a Gazebo simulation with an empty world and default physics.

    Args:
        step_size: Gazebo step size.
        real_time_factor: The desired real time factor of the simulation.
        steps_per_run: How many steps should be executed at each Gazebo run.

    Raises:
        RuntimeError: If the initialization of either the simulator or the world fails.

    Returns:
        * **gazebo** -- The Gazebo simulator.
        * **world** -- The default world.
    """

    # Create the simulator
    gazebo = scenario.GazeboSimulator(step_size, real_time_factor, steps_per_run)

    # Initialize the simulator
    ok_initialize = gazebo.initialize()

    if not ok_initialize:
        raise RuntimeError("Failed to initialize Gazebo")

    # Get the world
    world = gazebo.get_world()

    # Insert the ground plane
    ok_ground = world.insert_model(gym_ignition_models.get_model_file("ground_plane"))

    if not ok_ground:
        raise RuntimeError("Failed to insert the ground plane")

    ok_physics = world.set_physics_engine(scenario.PhysicsEngine_dart)

    if not ok_physics:
        raise RuntimeError("Failed to insert the physics plugin")

    return gazebo, world


def get_joint_positions_space(
    model: scenario.Model, considered_joints: List[str] = None
) -> gym.spaces.Box:
    """
    Build a Box space from the joint position limits.

    Args:
        model: The model from which generating the joint space.
        considered_joints: List of considered joints. It is helpful to restrict the set
            of joints and to enforce a custom joint serialization.

    Returns:
        A box space created from the model's joint position limits.
    """

    if considered_joints is None:
        considered_joints = model.joint_names()

    # Get the joint limits
    joint_limits = model.joint_limits(considered_joints)

    # Build the space
    space = gym.spaces.Box(
        low=np.array(joint_limits.min), high=np.array(joint_limits.max)
    )

    return space



# Copyright (C) 2019 Istituto Italiano di Tecnologia (IIT). All rights reserved.
# This software may be modified and distributed under the terms of the
# GNU Lesser General Public License v2.1 or any later version.

import gym
import pytest
from gym_ignition.utils import logger

# Set verbosity
logger.set_level(gym.logger.DEBUG)


def template_run_environment(env_name):
    logger.info(f"Testing environment '{env_name}'")
    env = gym.make(env_name)
    assert env, f"Failed to create '{env_name}' environment"

    observation = env.observation_space.sample()
    assert observation.size > 0, "The sampled observation is empty"

    observation = env.reset()
    assert observation.size > 0, "The observation is empty"

    for _ in range(10):
        action = env.action_space.sample()
        state, reward, done, _ = env.step(action)
        assert state.size > 0, "The environment didn't return a valid state"

    env.close()


@pytest.mark.parametrize(
    "env_name",
    [
        "CartPoleDiscrete-Gazebo-v0",
        "CartPoleContinuous-Gazebo-v0",
        "Pendulum-Gazebo-v0",
    ],
)
def test_environment(env_name: str):
    template_run_environment(env_name)



# Copyright (C) 2020 Istituto Italiano di Tecnologia (IIT). All rights reserved.
# This software may be modified and distributed under the terms of the
# GNU Lesser General Public License v2.1 or any later version.

from numbers import Number
from typing import List, Union

import numpy as np

from scenario import gazebo as scenario


def normalize(
    input: Union[Number, List[Number], np.ndarray],
    low: Union[Number, List[Number], np.ndarray],
    high: Union[Number, List[Number], np.ndarray],
) -> Union[Number, np.ndarray]:

    if low is None or high is None:
        return input

    if isinstance(input, Number):
        input = [input]

    if isinstance(low, Number):
        low = [low]

    if isinstance(high, Number):
        high = [high]

    output = scenario.normalize(list(input), list(low), list(high))

    if len(output) == 1:
        return output[0]
    else:
        return np.array(output)


def denormalize(
    input: Union[Number, List[Number], np.ndarray],
    low: Union[Number, List[Number], np.ndarray],
    high: Union[Number, List[Number], np.ndarray],
) -> Union[Number, np.ndarray]:

    if low is None or high is None:
        return input

    if isinstance(input, Number):
        input = [input]

    if isinstance(low, Number):
        low = [low]

    if isinstance(high, Number):
        high = [high]

    output = scenario.denormalize(list(input), list(low), list(high))

    if len(output) == 1:
        return output[0]
    else:
        return np.array(output)



# Copyright (C) 2019 Istituto Italiano di Tecnologia (IIT). All rights reserved.
# This software may be modified and distributed under the terms of the
# GNU Lesser General Public License v2.1 or any later version.

import abc

import gym
from gym_ignition.base.task import Task


class Runtime(gym.Env, abc.ABC):
    """
    Base class for defining executors of :py:class:`~gym_ignition.base.task.Task` objects.

    :py:class:`~gym_ignition.base.task.Task` classes are supposed to be generic and are
    not tied to any specific runtime. Implementations of a runtime class contain all the
    logic to define how to execute the task, allowing to run the same
    :py:class:`~gym_ignition.base.task.Task` class on different simulators or in a
    real-time setting.

    Runtimes are the real :py:class:`gym.Env` objects returned to the users when they call
    the :py:class:`gym.make` factory method.

    Args:
        task: the :py:class:`~gym_ignition.base.task.Task` object to handle.
        agent_rate: the rate at which the environment will be called. Sometimes tasks need
            to know this information.

    Example:
        Here is minimal example of how the :py:class:`Runtime`, :py:class:`gym.Env` and
        :py:class:`~gym_ignition.base.task.Task` could be integrated:

    .. code-block:: python

        class FooRuntime(Runtime):

            def __init__(self, task):
                super().__init__(task=task, agent_rate=agent_rate)
                self.action_space, self.observation_space = self.task.create_spaces()

            def reset(self):
                self.task.reset_task()
                return self.task.get_observation()

            def step(self, action):
                self.task.set_action(action)

                # [...] code that performs the real step [...]

                done = self.task.is_done()
                reward = self.task.get_reward()
                observation = self.task.get_observation()

                return observation, reward, done, {}

            def close(self):
                pass

    Note:
        Runtimes can handle only one :py:class:`~gym_ignition.base.task.Task` object.
    """

    def __init__(self, task: Task, agent_rate: float):

        #: Task handled by the runtime.
        self.task: Task = task

        #: Rate of environment execution.
        self.agent_rate = agent_rate

    @abc.abstractmethod
    def timestamp(self) -> float:
        """
        Return the timestamp associated to the execution of the environment.

        In real-time environments, the timestamp is the time read from the host system.
        In simulated environments, the timestamp is the simulated time, which might not
        match the real-time in the case of a real-time factor different than 1.

        Returns:
            The current environment timestamp.
        """



# Copyright (C) 2019 Istituto Italiano di Tecnologia (IIT). All rights reserved.
# This software may be modified and distributed under the terms of the
# GNU Lesser General Public License v2.1 or any later version.

import abc
from typing import Tuple

import gym
import numpy as np
from gym_ignition.base import task
from gym_ignition.utils.typing import (
    Action,
    ActionSpace,
    Observation,
    ObservationSpace,
    Reward,
)

from scenario import core as scenario


class PendulumSwingUp(task.Task, abc.ABC):
    def __init__(self, agent_rate: float, **kwargs):

        # Initialize the Task base class
        task.Task.__init__(self, agent_rate=agent_rate)

        # Name of the pendulum model
        self.model_name = None

        # Limits
        self._max_speed = 10.0
        self._max_torque = 50.0

    def create_spaces(self) -> Tuple[ActionSpace, ObservationSpace]:

        action_space = gym.spaces.Box(
            low=-self._max_torque, high=self._max_torque, shape=(1,), dtype=np.float32
        )

        high = np.array([1.0, 1.0, self._max_speed])  # cos(theta)  # sin(theta)

        observation_space = gym.spaces.Box(low=-high, high=high, dtype=np.float32)

        return action_space, observation_space

    def set_action(self, action: Action) -> None:

        # Get the force value
        force = action.tolist()[0]

        # Set the force value
        model = self.world.get_model(self.model_name)
        ok_force = model.get_joint("pivot").set_generalized_force_target(force)

        if not ok_force:
            raise RuntimeError("Failed to set the force to the pendulum")

    def get_observation(self) -> Observation:

        # Get the model
        model = self.world.get_model(self.model_name)

        # Get the new joint position and velocity
        q = model.get_joint("pivot").position()
        dq = model.get_joint("pivot").velocity()

        # Create the observation
        observation = Observation(np.array([np.cos(q), np.sin(q), dq]))

        # Return the observation
        return observation

    def get_reward(self) -> Reward:

        # This environment is done only if the observation goes outside its limits.
        # Since it can happen only when velocity is too high, penalize this happening.
        cost = 100.0 if self.is_done() else 0.0

        # Get the model
        model = self.world.get_model(self.model_name)

        # Get the pendulum state
        q = model.get_joint("pivot").position()
        dq = model.get_joint("pivot").velocity()
        tau = model.get_joint("pivot").generalized_force_target()

        # Calculate the cost
        cost += (q ** 2) + 0.1 * (dq ** 2) + 0.001 * (tau ** 2)

        return Reward(-cost)

    def is_done(self) -> bool:

        # Get the observation
        observation = self.get_observation()

        # The environment is done if the observation is outside its space
        done = not self.observation_space.contains(observation)

        return done

    def reset_task(self) -> None:

        if self.model_name not in self.world.model_names():
            raise RuntimeError("The cartpole model was not inserted in the world")

        # Get the model
        model = self.world.get_model(self.model_name)

        # Control the pendulum in force mode
        pivot = model.get_joint("pivot")
        ok_control_mode = pivot.set_control_mode(scenario.JointControlMode_force)

        if not ok_control_mode:
            raise RuntimeError("Failed to change the control mode of the pendulum")

        # Sample an observation
        cos_q, sin_q, dq = self.observation_space.sample()

        # Compute the angle
        q = np.arctan2(sin_q, cos_q)

        # Convert to float
        q, dq = float(q), float(dq)

        # Reset the pendulum state
        ok_reset = pivot.to_gazebo().reset(q, dq)

        if not ok_reset:
            raise RuntimeError("Failed to reset the pendulum state")



# Copyright (C) 2019 Istituto Italiano di Tecnologia (IIT). All rights reserved.
# This software may be modified and distributed under the terms of the
# GNU Lesser General Public License v2.1 or any later version.

from gym_ignition.base import runtime, task
from gym_ignition.utils.typing import Action, Done, Info, Observation, State


class RealTimeRuntime(runtime.Runtime):
    """
    Implementation of :py:class:`~gym_ignition.base.runtime.Runtime` for real-time
    execution.

    Warning:
        This class is not yet complete.
    """

    def __init__(self, task_cls: type, robot_cls: type, agent_rate: float, **kwargs):

        # Build the environment
        task_object = task_cls(**kwargs)

        assert isinstance(
            task_object, task.Task
        ), "'task_cls' object must inherit from Task"

        super().__init__(task=task_object, agent_rate=agent_rate)

        raise NotImplementedError

    # =================
    # Runtime interface
    # =================

    def timestamp(self) -> float:

        raise NotImplementedError

    # =================
    # gym.Env interface
    # =================

    def step(self, action: Action) -> State:

        # Validate action and robot
        assert self.action_space.contains(action), "%r (%s) invalid" % (
            action,
            type(action),
        )

        # Set the action
        ok_action = self.task.set_action(action)
        assert ok_action, "Failed to set the action"

        # TODO: realtime step

        # Get the observation
        observation = self.task.get_observation()
        assert self.observation_space.contains(observation), "%r (%s) invalid" % (
            observation,
            type(observation),
        )

        # Get the reward
        reward = self.task.get_reward()
        assert reward, "Failed to get the reward"

        # Check termination
        done = self.task.is_done()

        return State((observation, reward, Done(done), Info({})))

    def reset(self) -> Observation:

        # Get the observation
        observation = self.task.get_observation()

        return Observation(observation)

    def render(self, mode: str = "human", **kwargs) -> None:
        raise NotImplementedError

    def close(self) -> None:
        raise NotImplementedError



# Copyright (C) 2020 Istituto Italiano di Tecnologia (IIT). All rights reserved.
# This software may be modified and distributed under the terms of the
# GNU Lesser General Public License v2.1 or any later version.

import abc


class ModelWithFile(abc.ABC):
    def __init__(self):

        super().__init__()

    @classmethod
    @abc.abstractmethod
    def get_model_file(cls) -> str:
        pass
